import functools
from datetime import timedelta, datetime
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_table
from dash.dependencies import Output, Input
from sqlalchemy.orm import Session
from covigator.dashboard.figures.variants import VariantsFigures
from covigator.dashboard.tabs import MONTH_PATTERN
from covigator.database.model import DataSource
from covigator.database.queries import Queries

ID_DROPDOWN_DATE_RANGE_END_DIV = 'dropdown-date-range-end-div'
ID_DROPDOWN_GENE = 'dropdown-gene'
ID_SLIDER_MIN_SAMPLES = 'slider-min-samples'
ID_SLIDER_MIN_COOCCURRENCES = 'slider-min-cooccurrences'
ID_DROPDOWN_SIMILARITY_METRIC = 'dropdown-heatmap-metric'
ID_SLIDER_BIN_SIZE = 'slider-bin-size'
BIN_SIZE_VALUES = [10, 50, 100, 200, 300, 400]
ID_DROPDOWN_DATE_RANGE_END = 'dropdown-date-range-end'
ID_DROPDOWN_DATE_RANGE_START = 'dropdown-date-range-start'
ID_TOP_VARIANTS_METRIC = 'dropdown-top-variants-metric'
ID_SLIDER_TOP_VARIANTS = 'slider-top-variants'
ID_VARIANTS_MDS = 'variants-mds'
ID_COOCCURRENCE_HEATMAP = 'cooccurrence-heatmap'
ID_NEEDLE_PLOT = 'needle-plot'
ID_TOP_OCCURRING_VARIANTS = 'top-occurring-variants'
ID_TOP_OCCURRING_VARIANTS_TABLE = 'top-occurring-variants-table'
ID_DROPDOWN_DATA_SOURCE = "dropdown-data-source-variants-tab"


@functools.lru_cache()
def get_tab_variants(queries: Queries):

    return dbc.Card(
        dbc.CardBody(
            children=[
                get_variants_tab_left_bar(queries=queries),
                get_variants_tab_graphs()
            ])
    )


def get_variants_tab_graphs():
    return html.Div(children=[
        html.Br(),
        html.Div(id=ID_TOP_OCCURRING_VARIANTS, children=dash_table.DataTable(id=ID_TOP_OCCURRING_VARIANTS_TABLE)),
        html.Br(),
        html.Div(id=ID_NEEDLE_PLOT),
        html.Br(),
        html.Div(id=ID_COOCCURRENCE_HEATMAP),
        html.Br(),
        html.Div(id=ID_VARIANTS_MDS),
        html.Br(),
    ], className="ten columns", style={'overflow': 'scroll', "height": "900px"}, )


def get_variants_tab_left_bar(queries: Queries):

    genes = queries.get_genes()
    months = queries.get_sample_months(MONTH_PATTERN)
    today = datetime.now()
    today_formatted = today.strftime(MONTH_PATTERN)
    oneyearago = today - timedelta(days=356)
    oneyearago_formatted = oneyearago.strftime(MONTH_PATTERN)

    return html.Div(children=[
        html.Br(),
        dcc.Markdown("""Select data source"""),
        dcc.Dropdown(
            id=ID_DROPDOWN_DATA_SOURCE,
            options=[{'label': DataSource.ENA.name, 'value': DataSource.ENA.name},
                     {'label': DataSource.GISAID.name, 'value': DataSource.GISAID.name}],
            value=None,
            multi=False
        ),
        html.Br(),
        dcc.Markdown("Select a gene"),
        dcc.Dropdown(
            id=ID_DROPDOWN_GENE,
            options=[{'label': c.name, 'value': c.name} for c in genes],
            value=None,
            multi=False
        ),
        html.Br(),
        dcc.Markdown("""**Top occurring variants**

Number of top occurring variants"""),
        dcc.Slider(
            id=ID_SLIDER_TOP_VARIANTS,
            min=10,
            max=100,
            step=10,
            value=10,
            dots=True,
            marks={i: '{}'.format(i) for i in [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]},
            tooltip=dict(always_visible=False, placement="right")
        ),
        html.Br(),
        dcc.Markdown("""Metric to measure abundance of variants per month"""),
        dcc.Dropdown(
            id=ID_TOP_VARIANTS_METRIC,
            options=[
                {'label': 'Count', 'value': 'count'},
                {'label': 'Frequency', 'value': 'frequency_by_month'}],
            value='count',
            clearable=False,
            multi=False
        ),
        html.Br(),
        dcc.Markdown("""Select a start and end date"""),
        html.Div(children=[
            dcc.Dropdown(
                id=ID_DROPDOWN_DATE_RANGE_START,
                options=[{'label': c, 'value': c} for c in months],
                value=oneyearago_formatted,
                multi=False,
                clearable=False
            ),
            html.Div(
                id=ID_DROPDOWN_DATE_RANGE_END_DIV,
                children=dcc.Dropdown(
                    id=ID_DROPDOWN_DATE_RANGE_END,
                    options=[{'label': c, 'value': c} for c in months],
                    value=today_formatted,
                    multi=False,
                    clearable=False
                ))]),
        html.Br(),
        dcc.Markdown("""**Genome view**
        
Bin size"""),
        dcc.Dropdown(
            id=ID_SLIDER_BIN_SIZE,
            options=[{'label': "{} bp".format(c), 'value': str(c)} for c in BIN_SIZE_VALUES],
            value=50,
            multi=False,
            clearable=False
        ),
        html.Br(),
        dcc.Markdown("""**Co-occurrence matrix**
        
Metric to assess paiwise co-occurrence"""),
        dcc.Dropdown(
            id=ID_DROPDOWN_SIMILARITY_METRIC,
            options=[{'label': "Count", 'value': "count"},
                     {'label': "Frequency", 'value': "frequency"},
                     {'label': "Jaccard index", 'value': "jaccard"},
                     {'label': "Cohen's kappa coefficient", 'value': "kappa"},
                     ],
            value="jaccard",
            clearable=False,
            multi=False
        ),
        dcc.Markdown("""Minimum number of pairwise co-occurrences"""),
        dcc.Slider(
            id=ID_SLIDER_MIN_COOCCURRENCES,
            min=1,
            max=100,
            step=5,
            value=10,
            dots=True,
            marks={i: '{}'.format(i) for i in [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]},
            tooltip=dict(always_visible=False, placement="right")
        ),
        html.Br(),
        dcc.Markdown("""**Variants clustering**
        
The number of samples (or total weight) in a neighborhood for a point to be 
considered as a core point. This includes the point itself."""),
        dcc.Slider(
            id=ID_SLIDER_MIN_SAMPLES,
            min=2,
            max=10,
            step=1,
            value=5,
            dots=True,
            marks={i: str(i) for i in range(2, 10)},
            tooltip=dict(always_visible=False, placement="right")
        ),
    ], className="two columns")


def set_callbacks_variants_tab(app, session: Session):

    queries = Queries(session=session)
    figures = VariantsFigures(queries=queries)

    @app.callback(
        Output(ID_TOP_OCCURRING_VARIANTS, 'children'),
        Input(ID_SLIDER_TOP_VARIANTS, 'value'),
        Input(ID_DROPDOWN_GENE, 'value'),
        Input(ID_DROPDOWN_DATE_RANGE_START, 'value'),
        Input(ID_DROPDOWN_DATE_RANGE_END, 'value'),
        Input(ID_TOP_VARIANTS_METRIC, 'value'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
    )
    def update_top_occurring_variants(top_variants, gene_name, date_range_start, date_range_end, metric, source):
        return html.Div(children=figures.get_top_occurring_variants_plot(
            top=top_variants, gene_name=gene_name, date_range_start=date_range_start,
            date_range_end=date_range_end, metric=metric, source=source))

    @app.callback(
        Output(ID_NEEDLE_PLOT, 'children'),
        Input(ID_DROPDOWN_GENE, 'value'),
        Input(ID_TOP_OCCURRING_VARIANTS_TABLE, "derived_virtual_data"),
        Input(ID_TOP_OCCURRING_VARIANTS_TABLE, "derived_virtual_selected_rows"),
        Input(ID_SLIDER_BIN_SIZE, 'value'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
    )
    def update_needle_plot(gene_name, rows, selected_rows_indices, bin_size, source):
        if gene_name is not None:
            selected_rows = [rows[s] for s in selected_rows_indices] if selected_rows_indices else None
            plot = html.Div(
                children=figures.get_variants_plot(
                    gene_name=gene_name,
                    selected_variants=selected_rows,
                    bin_size=int(bin_size),
                    source=source
                ))
        else:
            plot = html.Div(
                children=figures.get_variants_abundance_plot(bin_size=int(bin_size), source=source))
        return plot

    @app.callback(
        Output(ID_DROPDOWN_DATE_RANGE_END_DIV, 'children'),
        Input(ID_DROPDOWN_DATE_RANGE_START, 'value'),
    )
    def update_dropdown_end_date(start_date):
        today = datetime.now()
        today_formatted = today.strftime(MONTH_PATTERN)
        months = [m for m in queries.get_sample_months(MONTH_PATTERN) if m >= start_date]
        return dcc.Dropdown(
            id=ID_DROPDOWN_DATE_RANGE_END,
            options=[{'label': c, 'value': c} for c in months],
            value=today_formatted,
            multi=False,
            clearable=False
        )

    @app.callback(
        Output(ID_COOCCURRENCE_HEATMAP, 'children'),
        Input(ID_DROPDOWN_GENE, 'value'),
        Input(ID_TOP_OCCURRING_VARIANTS_TABLE, "derived_virtual_data"),
        Input(ID_TOP_OCCURRING_VARIANTS_TABLE, "derived_virtual_selected_rows"),
        Input(ID_DROPDOWN_SIMILARITY_METRIC, 'value'),
        Input(ID_SLIDER_MIN_COOCCURRENCES, 'value'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
    )
    def update_cooccurrence_heatmap(gene_name, rows, selected_rows_indices, metric, min_occurrences, source):
        if source != DataSource.ENA.name:
            plot = html.Div(
                children=[dcc.Markdown(
                    """**The cooccurrence heatmap is currently only available for the ENA dataset**""")])
        elif gene_name is None:
            plot = html.Div(
                children=[dcc.Markdown(
                    """**Please, select a gene to explore the cooccurrence heatmap**""")])
        else:
            selected_rows = [rows[s] for s in selected_rows_indices] if selected_rows_indices else None
            plot = html.Div(children=figures.get_cooccurrence_heatmap(
                gene_name=gene_name,
                selected_variants=selected_rows,
                metric=metric,
                min_occurrences=min_occurrences))
        return plot

    @app.callback(
        Output(ID_VARIANTS_MDS, 'children'),
        Input(ID_DROPDOWN_GENE, 'value'),
        Input(ID_TOP_OCCURRING_VARIANTS_TABLE, "derived_virtual_data"),
        Input(ID_TOP_OCCURRING_VARIANTS_TABLE, "derived_virtual_selected_rows"),
        Input(ID_SLIDER_MIN_COOCCURRENCES, 'value'),
        Input(ID_SLIDER_MIN_SAMPLES, 'value'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
    )
    def update_variants_mds(gene_name, rows, selected_rows_indices, min_cooccurrence, min_samples, source):
        if source != DataSource.ENA.name:
            plot = html.Div(children=
                            [dcc.Markdown(
                                """**The variants clustering is currently only available for the ENA dataset**""")])
        elif gene_name is None:
            plot = html.Div(
                children=[dcc.Markdown(
                    """**Please, select a gene to explore the cooccurrence clusters**""")])
        else:
            selected_rows = [rows[s] for s in selected_rows_indices] if selected_rows_indices else None
            plot = html.Div(children=figures.get_variants_clustering(
                gene_name=gene_name,
                selected_variants=selected_rows,
                min_cooccurrence=min_cooccurrence,
                min_samples=min_samples))
            #plot = None
        return plot


import functools

import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html

from covigator.dashboard.tabs import get_mini_container, print_number
from covigator.database.queries import Queries


@functools.lru_cache()
def get_tab_overview(queries: Queries):

    count_samples = queries.count_samples()
    count_countries = queries.count_countries()
    count_variants = queries.count_variants()

    return dbc.Card(
        dbc.CardBody([
            get_header(),
            html.Div(
               children=[
                   html.P(
                       "Human infections with SARS-CoV-2 are spreading globally since the beginning of 2020, "
                       "necessitating preventive or therapeutic strategies and first steps towards an end to "
                       "this pandemic were done with the approval of the first mRNA vaccines against SARS-CoV-2. "
                       "We want to provide an interactive view on different types of mutations that can be "
                       "incorporated in global efforts to sustainably prevent or treat infections. Thus, we "
                       "envision to help guiding global vaccine design efforts to overcome the threats of this "
                       "pandemic."),
                   dcc.Markdown("""
                       CoVigator loads publicly available SARS-CoV-2 DNA sequences from two systems: 

                       * European Nucleotide Archive (ENA) providing raw reads in FASTQ format
                       * Global Initiative on Sharing Avian Influenza Data (GISAID) providing assemblies in FASTA format
                       """),
                   html.Br(),
                   html.Div(
                       html.Span(
                           children=[
                               get_mini_container(
                                   title="Samples",
                                   value=print_number(count_samples)
                               ),
                               get_mini_container(
                                   title="Countries",
                                   value=print_number(count_countries)
                               ),
                               get_mini_container(
                                   title="Mutations",
                                   value=print_number(count_variants)
                               )
                       ])
                   ),
                   html.Br(),
                   html.P("If you want to cite us:"),
                   html.P([
                       "Schrörs, B., Gudimella, R., Bukur, T., Rösler, T., Löwer, M., & Sahin, U. (2021). "
                       "Large-scale analysis of SARS-CoV-2 spike-glycoprotein mutants demonstrates the "
                       "need for continuous screening of virus isolates. BioRxiv, 2021.02.04.429765. ",
                       ],
                       style={"font-style": "italic", "margin-left": "50px"}),
                   html.P(
                       html.A("https://doi.org/10.1101/2021.02.04.429765",
                              href="https://doi.org/10.1101/2021.02.04.429765"),
                       style={"text-indent": "50px"}
                   ),

                   html.Br(),
                   html.P(["The CoVigator analysis pipeline processes SARS-CoV-2 FASTQ or FASTA files into "
                           "annotated and normalized analysis ready VCF files. The pipeline is implemented "
                           "in the Nextflow framework (Di Tommaso, 2017). "
                           "It is open sourced under the MIT license, see the repository for "
                           "more details ",
                           html.A("https://github.com/TRON-Bioinformatics/covigator-ngs-pipeline",
                                  href="https://github.com/TRON-Bioinformatics/covigator-ngs-pipeline"),
                           "."]),
                   html.Br(),
                   html.P("If you would like to cite the pipeline:"),
                   html.P([
                       "Riesgo-Ferreiro, P., Sorn, P., & Bukur, T.. (2021, July 8). "
                       "TRON-Bioinformatics/covigator-ngs-pipeline: Release v0.5.0 (Version v0.5.0). "
                       "Zenodo. ",

                   ], style={"font-style": "italic", "margin-left": "50px"}),
                   html.P(
                       html.A("http://doi.org/10.5281/zenodo.5082444",
                              href="http://doi.org/10.5281/zenodo.5082444"),
                       style={"text-indent": "50px"}
                   ),
                   html.Br(),
               ],
               style={"text-align": "left", "font-size": 16}),
        ]))


def get_header():
    logo = "/assets/CoVigator_logo_txt_reg_no_bg.png"
    return html.Div(
            [
                html.Div(
                    [html.Img(src=logo, id="covigator-logo",
                              style={"height": "100px", "width": "auto", "margin-bottom": "10px"},)
                    ],
                ),
                html.Div(className="one column"),
                html.Div(
                    [html.Div([html.H1("Monitoring SARS-CoV-2 mutations")], style={"text-align": "left"})],
                    id="title",
                ),
            ],
            id="header",
            className="row flex-display",
        )



from datetime import date, datetime
from sqlalchemy.orm import Session

from covigator.accessor import MINIMUM_DATE
from covigator.database.model import SampleGisaid, JobGisaid, Sample, DataSource, Log, CovigatorModule
from covigator.database.database import Database
from logzero import logger
from Bio import SeqIO
import time

from covigator.exceptions import CovigatorExcludedSampleTooEarlyDateException
from covigator.misc.compression import compress_sequence
from covigator.misc.country_parser import CountryParser

THRESHOLD_NON_VALID_BASES_RATIO = 0.2
THRESHOLD_GENOME_COVERAGE = 0.2
CHROMOSOME = "MN908947.3"
GENOME_LENGTH = 29903
BATCH_SIZE = 1000


class GisaidAccessor:

    def __init__(self, input_fasta: str, input_metadata: str, database: Database):
        self.start_time = datetime.now()
        self.input_fasta = input_fasta
        self.input_metadata = input_metadata
        self.has_error = False

        self.database = database
        assert self.database is not None, "Empty database"

        self.excluded_by_host = 0
        self.excluded_by_horizontal_coverage = 0
        self.excluded_by_non_valid_bases_ratio = 0
        self.excluded_existing = 0
        self.excluded_by_date = 0
        self.included = 0
        self.country_parser = CountryParser()

    def access(self):
        session = self.database.get_database_session()
        # NOTE: holding in memory the whole list of existing ids is much faster than querying every time
        # it assumes there will be no repetitions
        existing_sample_ids = [value for value, in session.query(SampleGisaid.run_accession).all()]
        try:
            num_samples = self.process(existing_sample_ids, session)
            logger.info("Read file of {} GISAID samples".format(num_samples))
        except Exception as e:
            logger.exception(e)
            session.rollback()
            self.has_error = True
        finally:
            self._write_execution_log(session)
            session.close()
            self._log_results()
            logger.info("Finished GISAID accessor")

    def process(self, existing_sample_ids, session) -> int:
        meta_dict = {}
        logger.info("Reading metadata file...")
        with open(self.input_metadata) as metadata:
            for line in metadata:
                elements = line.rstrip().split("\t")
                virus_name = elements[0]
                virus_type = elements[1]
                accession_id = elements[2]
                collection_date = elements[3]
                location = elements[4]
                host = elements[7]
                meta_dict[virus_name] = (virus_type, accession_id, collection_date, location, host)
        logger.info("Metadata (num_samples={}) loaded.".format(len(meta_dict)))

        num_samples = 0
        gisaid_samples = set()
        total_time = 0
        logger.info("Reading FASTA...")
        samples_gisaid = []
        jobs_and_samples = []
        for record in SeqIO.parse(self.input_fasta, "fasta"):
            start = time.time()

            fields = record.description.split("|")
            identifier = fields[0]

            if identifier in gisaid_samples:
                continue
            gisaid_samples.add(identifier)

            if identifier in existing_sample_ids:
                num_samples += 1
                # we skip samples already in the database
                self.excluded_existing += 1
                continue

            metadata = meta_dict.get(identifier)
            if not metadata:
                continue
            
            location_list = metadata[3].split("/")
            country_raw = "None"
            try:
                country_raw = location_list[1].strip()
            except:
                pass
            region = "None"
            try:
                region = location_list[2].strip()
            except:
                pass

            host = metadata[4].lower().strip()
            if host != "human":
                num_samples += 1
                self.excluded_by_host += 1
                continue

            sequence_length = len(record.seq)
            if float(sequence_length) / GENOME_LENGTH < THRESHOLD_GENOME_COVERAGE:
                num_samples += 1
                self.excluded_by_horizontal_coverage += 1
                continue

            count_n_bases = record.seq.count("N")
            count_ambiguous_bases = sum([record.seq.count(b) for b in "RYWSMKHBVD"])
            if float(count_n_bases + count_ambiguous_bases) / sequence_length > THRESHOLD_NON_VALID_BASES_RATIO:
                num_samples += 1
                self.excluded_by_non_valid_bases_ratio += 1
                continue

            sample_gisaid = SampleGisaid(
                run_accession=identifier,
                date=metadata[2],
                host_tax_id=None,
                host=host,
                country_raw=country_raw,
                region=region,
                country=None,
                country_alpha_2=None,
                country_alpha_3=None,
                continent=None,
                continent_alpha_2=None,
                site=None,
                site2=None,
                sequence={CHROMOSOME: compress_sequence(record.seq)},
                sequence_length=sequence_length,
                count_n_bases=count_n_bases,
                count_ambiguous_bases=count_ambiguous_bases
            )
            try:
                self._parse_country(sample_gisaid)
                self._parse_dates(sample_gisaid)
                sample = self._build_sample(sample_gisaid)
                job = JobGisaid(run_accession=sample_gisaid.run_accession)
                samples_gisaid.append(sample_gisaid)
                jobs_and_samples.append(job)
                jobs_and_samples.append(sample)
                num_samples += 1
                self.included += 1
                end = time.time()
                total_time += end - start

                if len(samples_gisaid) == BATCH_SIZE:
                    session.add_all(samples_gisaid)
                    session.commit()
                    session.add_all(jobs_and_samples)
                    session.commit()
                    samples_gisaid = []
                    jobs_and_samples = []
            except CovigatorExcludedSampleTooEarlyDateException:
                logger.error("Sample excluded due to too early date")
                self.excluded_by_date += 0

        if len(samples_gisaid) > 0:
            session.add_all(samples_gisaid)
            session.commit()
            session.add_all(jobs_and_samples)
            session.commit()
        if num_samples > 0:
            logger.info("It took {} secs/sample on average".format(float(total_time) / num_samples))
        return num_samples

    def _build_sample(self, sample_gisaid):
        return Sample(
            id=sample_gisaid.run_accession,
            source=DataSource.GISAID,
            gisaid_id=sample_gisaid.run_accession
        )

    def _parse_country(self, gisaid_sample: SampleGisaid):
        parsed_country = self.country_parser.parse_country(gisaid_sample.country_raw)
        gisaid_sample.country = parsed_country.country
        gisaid_sample.country_alpha_2 = parsed_country.country_alpha_2
        gisaid_sample.country_alpha_3 = parsed_country.country_alpha_3
        gisaid_sample.continent_alpha_2 = parsed_country.continent_alpha_2
        gisaid_sample.continent = parsed_country.continent

    def _parse_dates(self, gisaid_sample: SampleGisaid):
        gisaid_sample.date = self._parse_abstract(gisaid_sample.date, date.fromisoformat)
        if gisaid_sample.date is not None and gisaid_sample.date < MINIMUM_DATE:
            raise CovigatorExcludedSampleTooEarlyDateException

    def _parse_abstract(self, value, type):
        try:
            value = type(value)
        except (ValueError, TypeError):
            value = None
        return value

    def _log_results(self):
        logger.info("Included new runs = {}".format(self.included))
        logger.info("Excluded already existing runs = {}".format(self.excluded_existing))
        logger.info("Excluded non human host = {}".format(self.excluded_by_host))
        logger.info("Excluded by low horizontal coverage = {}".format(self.excluded_by_horizontal_coverage))
        logger.info("Excluded by high non valid bases ratio = {}".format(self.excluded_by_non_valid_bases_ratio))

    def _write_execution_log(self, session: Session):
        end_time = datetime.now()
        session.add(Log(
            start=self.start_time,
            end=end_time,
            source=DataSource.GISAID,
            module=CovigatorModule.ACCESSOR,
            has_error=self.has_error,
            data={
                "included": self.included,
                "excluded": {
                    "existing": self.excluded_existing,
                    "excluded_by_host": self.excluded_by_host,
                    "excluded_by_date": self.excluded_by_date,
                    "excluded_by_horizontal_coverage": self.excluded_by_horizontal_coverage,
                    "excluded_by_non_valid_bases_ratio": self.excluded_by_non_valid_bases_ratio
                }
            }
        ))
        session.commit()





class CovigatorDatabaseConnectionException(Exception):
    pass


class CovigatorPipelineError(Exception):
    pass


class CovigatorDashBoardInitialisationError(Exception):
    pass


class CovigatorExcludedAssemblySequence(Exception):
    pass


class CovigatorNotSupportedVariant(Exception):
    pass


class CovigatorQueryException(Exception):
    pass


class CovigatorExcludedSampleTooEarlyDateException(Exception):
    pass


class CovigatorExcludedSampleTooManyMutations(Exception):
    pass


class CovigatorErrorProcessingCoverageResults(Exception):
    pass


class CovigatorExcludedSampleNarrowCoverage(Exception):
    pass


class CovigatorExcludedSampleBadQualityReads(Exception):
    pass


class CovigatorDashboardMissingPrecomputedData(Exception):
    pass


import os
import pkg_resources
import covigator
from covigator.configuration import Configuration


class FakeConfiguration(Configuration):

    def __init__(self):
        # use a folder with write permissions for testing
        os.environ[self.ENV_COVIGATOR_STORAGE_FOLDER] = "./data/covigator"
        os.environ[self.ENV_COVIGATOR_REF_FASTA] = pkg_resources.resource_filename(
            covigator.tests.__name__, "resources/MN908947.3.fa")
        os.environ[self.ENV_COVIGATOR_GENE_ANNOTATIONS] = pkg_resources.resource_filename(
            covigator.tests.__name__, "resources/sars_cov_2.json")
        super().__init__()



from datetime import datetime
from typing import List
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Float, Enum, DateTime, Integer, Boolean, Date, ForeignKey, \
    ForeignKeyConstraint, BigInteger, JSON
import enum

from covigator.configuration import Configuration


def get_table_versioned_name(basename, config: Configuration):
    return "{}{}".format(basename, config.db_table_version)


config = Configuration()
GENE_TABLE_NAME = get_table_versioned_name('gene', config=config)
LOG_TABLE_NAME = get_table_versioned_name('log', config=config)
VARIANT_COOCCURRENCE_TABLE_NAME = get_table_versioned_name('variant_cooccurrence', config=config)
VARIANT_OBSERVATION_TABLE_NAME = get_table_versioned_name('variant_observation', config=config)
SUBCLONAL_VARIANT_OBSERVATION_TABLE_NAME = get_table_versioned_name('subclonal_variant_observation', config=config)
VARIANT_TABLE_NAME = get_table_versioned_name('variant', config=config)
JOB_ENA_TABLE_NAME = get_table_versioned_name('job_ena', config=config)
JOB_GISAID_TABLE_NAME = get_table_versioned_name('job_gisaid', config=config)
SAMPLE_TABLE_NAME = get_table_versioned_name('sample', config=config)
SAMPLE_GISAID_TABLE_NAME = get_table_versioned_name('sample_gisaid', config=config)
SAMPLE_ENA_TABLE_NAME = get_table_versioned_name('sample_ena', config=config)
CONSERVATION_TABLE_NAME = get_table_versioned_name('conservation', config=config)
PRECOMPUTED_VARIANTS_PER_SAMPLE_TABLE_NAME = get_table_versioned_name('precomputed_variants_per_sample', config=config)
PRECOMPUTED_SUBSTITUTIONS_COUNTS_TABLE_NAME = get_table_versioned_name('precomputed_substitutions_counts', config=config)
PRECOMPUTED_INDEL_LENGTH_TABLE_NAME = get_table_versioned_name('precomputed_indel_length', config=config)
PRECOMPUTED_ANNOTATION_TABLE_NAME = get_table_versioned_name('precomputed_annotation', config=config)
PRECOMPUTED_OCCURRENCE_TABLE_NAME = get_table_versioned_name('precomputed_top_occurrence', config=config)
PRECOMPUTED_DN_DS_TABLE_NAME = get_table_versioned_name('precomputed_dn_ds', config=config)
PRECOMPUTED_DN_DS_BY_DOMAIN_TABLE_NAME = get_table_versioned_name('precomputed_dn_ds_by_domain', config=config)
PRECOMPUTED_TABLE_COUNTS_TABLE_NAME = get_table_versioned_name('precomputed_table_counts', config=config)
PRECOMPUTED_VARIANT_ABUNDANCE_HIST_TABLE_NAME = get_table_versioned_name('precomputed_variant_abundance_histogram', config=config)
JOB_STATUS_CONSTRAINT_NAME = get_table_versioned_name('job_status', config=config)
DATA_SOURCE_CONSTRAINT_NAME = get_table_versioned_name('data_source', config=config)
COVIGATOR_MODULE_CONSTRAINT_NAME = get_table_versioned_name('covigator_module', config=config)
VARIANT_TYPE_CONSTRAINT_NAME = get_table_versioned_name('variant_type', config=config)
SEPARATOR = ";"

Base = declarative_base()


class Gene(Base):
    """
    This table holds the genes in the organism genome and its annotations in the field `data`. The annotations
    are in JSON format. This data is fetched from ftp://ftp.ensemblgenomes.org/pub/viruses/json/sars_cov_2/sars_cov_2.json
    """
    __tablename__ = GENE_TABLE_NAME

    identifier = Column(String, primary_key=True)
    name = Column(String)
    start = Column(Integer, index=True)
    end = Column(Integer)
    data = Column(JSON)

    def get_pfam_domains(self):
        protein_features = self.data.get("transcripts", [])[0].get("translations", [])[0].get("protein_features")
        pfam_protein_features = [f for f in protein_features if f.get("dbname") == "Pfam"]
        return sorted(pfam_protein_features, key=lambda d: int(d.get("start")))


class JobStatus(enum.Enum):
    """
    Valid job status
    """
    __constraint_name__ = JOB_STATUS_CONSTRAINT_NAME

    PENDING = 1
    QUEUED = 2
    DOWNLOADED = 3
    PROCESSED = 4
    LOADED = 5
    FAILED_DOWNLOAD = 6
    FAILED_PROCESSING = 7
    FAILED_LOAD = 8
    COOCCURRENCE = 9
    FAILED_COOCCURRENCE = 10
    FINISHED = 11
    HOLD = 12
    EXCLUDED = 13


class DataSource(enum.Enum):
    """
    Valid sources of data
    """
    __constraint_name__ = DATA_SOURCE_CONSTRAINT_NAME

    ENA = 1
    GISAID = 2


class VariantType(enum.Enum):
    __constraint_name__ = VARIANT_TYPE_CONSTRAINT_NAME

    SNV = 1
    INSERTION = 2
    DELETION = 3


class SampleGisaid(Base):
    """
    The table that holds all metadata for a GISAID sample
    """
    __tablename__ = SAMPLE_GISAID_TABLE_NAME

    run_accession = Column(String, primary_key=True)
    finished = Column(Boolean)
    date = Column(Date)
    # Host information
    host_tax_id = Column(String)
    host = Column(String)
    # geographical data
    country_raw = Column(String)
    region = Column(String)
    country = Column(String)
    country_alpha_2 = Column(String)
    country_alpha_3 = Column(String)
    continent = Column(String)
    continent_alpha_2 = Column(String)
    site = Column(String)
    site2 = Column(String)

    # sequence  information
    sequence = Column(JSON)
    sequence_length = Column(Integer)
    count_n_bases = Column(Integer)
    count_ambiguous_bases = Column(Integer)

    # counts of variants
    count_snvs = Column(Integer)
    count_insertions = Column(Integer)
    count_deletions = Column(Integer)


class SampleEna(Base):
    """
    The table that holds all metadata for a ENA sample
    """
    __tablename__ = SAMPLE_ENA_TABLE_NAME

    # data on run
    # TODO: add foreign keys to jobs
    run_accession = Column(String, primary_key=True)            # 'ERR4080473',
    finished = Column(Boolean)
    sample_accession = Column(String)                           # 'SAMEA6798401',
    scientific_name = Column(String)
    study_accession = Column(String)
    experiment_accession = Column(String)
    # TODO: store these as dates
    first_created = Column(Date)
    collection_date = Column(Date)
    instrument_platform = Column(String)
    instrument_model = Column(String)
    library_name = Column(String)
    nominal_length = Column(Integer)
    library_layout = Column(String)
    library_strategy = Column(String)
    library_source = Column(String)
    library_selection = Column(String)
    read_count = Column(BigInteger)
    base_count = Column(BigInteger)
    sample_collection = Column(String)
    sequencing_method = Column(String)
    center_name = Column(String)

    # FASTQs
    fastq_ftp = Column(String, nullable=False)                  # 'ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080473/ERR4080473_1.fastq.gz',
    fastq_md5 = Column(String, nullable=False)
    num_fastqs = Column(Integer, nullable=False)

    # data on host
    host_tax_id = Column(String)                                # '9606',
    host_sex = Column(String)
    host_body_site = Column(String)
    host_gravidity = Column(String)
    host_phenotype = Column(String)
    host_genotype = Column(String)

    # geographical data
    lat = Column(Float)
    lon = Column(Float)
    country_raw = Column(String)                                    # 'Denmark',
    country = Column(String)
    country_alpha_2 = Column(String)
    country_alpha_3 = Column(String)
    continent = Column(String)
    continent_alpha_2 = Column(String)

    # counts of variants
    count_snvs = Column(Integer)
    count_insertions = Column(Integer)
    count_deletions = Column(Integer)
    count_subclonal_snvs = Column(Integer)
    count_subclonal_insertions = Column(Integer)
    count_subclonal_deletions = Column(Integer)

    def get_fastqs_ftp(self) -> List:
        return self.fastq_ftp.split(SEPARATOR) if self.fastq_ftp is not None else []

    def get_fastqs_md5(self) -> List:
        return self.fastq_md5.split(SEPARATOR) if self.fastq_md5 is not None else []


class Sample(Base):
    """
    This table holds all samples loaded into Covigator irrespective of the data source.
    The same sample may be loaded from different data sources.
    There are foreign keys fields pointing to the source-specific tables with all metadata for the sample.
    """
    __tablename__ = SAMPLE_TABLE_NAME

    id = Column(String, primary_key=True)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__), primary_key=True)
    # NOTE: should have only one filled, either ena_id or gisaid_id and be coherent with the value of source
    ena_id = Column(ForeignKey("{}.run_accession".format(SampleEna.__tablename__)))
    gisaid_id = Column(ForeignKey("{}.run_accession".format(SampleGisaid.__tablename__)))


class JobGisaid(Base):
    """
    The table that holds an GISAID job
    """
    __tablename__ = JOB_GISAID_TABLE_NAME

    run_accession = Column(ForeignKey("{}.run_accession".format(SampleGisaid.__tablename__)), primary_key=True)

    # job status
    status = Column(Enum(JobStatus, name=JobStatus.__constraint_name__), default=JobStatus.PENDING)
    created_at = Column(DateTime(timezone=True), nullable=False, default=datetime.now())
    queued_at = Column(DateTime(timezone=True))
    analysed_at = Column(DateTime(timezone=True))
    loaded_at = Column(DateTime(timezone=True))
    failed_at = Column(DateTime(timezone=True))
    error_message = Column(String)
    vcf_path = Column(String)


class JobEna(Base):
    """
    The table that holds an ENA job
    """
    __tablename__ = JOB_ENA_TABLE_NAME

    run_accession = Column(ForeignKey("{}.run_accession".format(SampleEna.__tablename__)), primary_key=True)

    # job status
    status = Column(Enum(JobStatus, name=JobStatus.__constraint_name__), default=JobStatus.PENDING)
    created_at = Column(DateTime(timezone=True), nullable=False, default=datetime.now())
    queued_at = Column(DateTime(timezone=True))
    downloaded_at = Column(DateTime(timezone=True))
    analysed_at = Column(DateTime(timezone=True))
    cleaned_at = Column(DateTime(timezone=True))
    loaded_at = Column(DateTime(timezone=True))
    cooccurrence_at = Column(DateTime(timezone=True))
    failed_at = Column(DateTime(timezone=True))
    error_message = Column(String)

    # local files storage
    fastq_path = Column(String)     # the local path where FASTQ files are stored in semi colon separated list
    vcf_path = Column(String)
    # FASTP results
    qc = Column(JSON)
    qc_path = Column(String)
    # coverage analysis results
    horizontal_coverage_path = Column(String)
    vertical_coverage_path = Column(String)
    num_reads = Column(Integer)
    covered_bases = Column(Integer)
    coverage = Column(Float)
    mean_depth = Column(Float)
    mean_base_quality = Column(Float)
    mean_mapping_quality = Column(Float)

    def get_fastq_paths(self):
        return self.fastq_path.split(SEPARATOR) if self.fastq_path is not None else []

    def get_fastq1_and_fastq2(self):
        fastqs = self.get_fastq_paths()
        if len(fastqs) > 1:
            fastq1 = next(filter(lambda x: "_1.fastq" in x, fastqs))
            fastq2 = next(filter(lambda x: "_2.fastq" in x, fastqs))
        else:
            fastq1 = fastqs[0]
            fastq2 = None
        return fastq1, fastq2


class Variant(Base):
    """
    A variant with its specific annotations. THis does not contain any sample specific annotations.
    """
    __tablename__ = VARIANT_TABLE_NAME

    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer, index=True)
    reference = Column(String)
    alternate = Column(String)
    overlaps_multiple_genes = Column(Boolean, default=False)
    """
    SnpEff annotations
    ##INFO=<ID=ANN,Number=.,Type=String,Description="Functional annotations: '
    Allele |                    C 
    Annotation |                missense_variant
    Annotation_Impact |         MODERATE 
    Gene_Name |                 orf1ab 
    Gene_ID |                   gene-GU280_gp01 
    Feature_Type |              transcript 
    Feature_ID |                TRANSCRIPT_gene-GU280_gp01 
    Transcript_BioType |        protein_coding 
    Rank |                      1/1 
    HGVS.c |                    c.33A>C 
    HGVS.p |                    p.Lys11Asn 
    cDNA.pos / cDNA.length |    33/21290 
    CDS.pos / CDS.length |      33/21290 
    AA.pos / AA.length |        11/7095 
    Distance | 
    ERRORS / WARNINGS / INFO'   
    """
    annotation = Column(String, index=True)
    annotation_highest_impact = Column(String, index=True)
    annotation_impact = Column(String, index=True)
    gene_name = Column(String, index=True)
    gene_id = Column(String)
    biotype = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)
    cdna_pos_length = Column(String)
    cds_pos_length = Column(String)
    aa_pos_length = Column(String)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    def get_variant_id(self):
        return "{}:{}>{}".format(self.position, self.reference, self.alternate)


class VariantObservation(Base):
    """
    A variant observation in a particular sample. This contains all annotations of a specific observation of a variant.
    """
    __tablename__ = VARIANT_OBSERVATION_TABLE_NAME

    source = Column(Enum(DataSource, name=DataSource.__constraint_name__), primary_key=True)
    sample = Column(String, primary_key=True)
    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer)
    reference = Column(String)
    alternate = Column(String)
    quality = Column(Float)
    filter = Column(String)
    """
    ##INFO=<ID=DP,Number=1,Type=Integer,Description="Raw Depth">
    ##INFO=<ID=AF,Number=1,Type=Float,Description="Allele Frequency">
    ##INFO=<ID=SB,Number=1,Type=Integer,Description="Phred-scaled strand bias at this position">
    ##INFO=<ID=DP4,Number=4,Type=Integer,Description="Counts for ref-forward bases, ref-reverse, alt-forward and alt-reverse bases">
    ##INFO=<ID=INDEL,Number=0,Type=Flag,Description="Indicates that the variant is an INDEL.">
    ##INFO=<ID=CONSVAR,Number=0,Type=Flag,Description="Indicates that the variant is a consensus variant (as opposed to a low frequency variant).">
    ##INFO=<ID=HRUN,Number=1,Type=Integer,Description="Homopolymer length to the right of report indel position">
    """
    dp = Column(Integer)
    dp4_ref_forward = Column(Integer)
    dp4_ref_reverse = Column(Integer)
    dp4_alt_forward = Column(Integer)
    dp4_alt_reverse = Column(Integer)
    vaf = Column(Float)
    strand_bias = Column(Integer)
    # fields replicated from Variant for performance reasons
    annotation = Column(String)
    annotation_impact = Column(String)
    biotype = Column(String)
    annotation_highest_impact = Column(String)
    gene_name = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)

    # fields replicated from sample for performance reasons
    date = Column(Date)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    ForeignKeyConstraint([sample, source], [Sample.id, Sample.source])
    ForeignKeyConstraint([variant_id], [Variant.variant_id])


class SubclonalVariantObservation(Base):
    """
    A variant observation in a particular sample. This contains all annotations of a specific observation of a variant.
    """
    __tablename__ = SUBCLONAL_VARIANT_OBSERVATION_TABLE_NAME

    source = Column(Enum(DataSource, name=DataSource.__constraint_name__), primary_key=True)
    sample = Column(String, primary_key=True)
    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer)
    reference = Column(String)
    alternate = Column(String)
    quality = Column(Float)
    filter = Column(String)
    """
    ##INFO=<ID=DP,Number=1,Type=Integer,Description="Raw Depth">
    ##INFO=<ID=AF,Number=1,Type=Float,Description="Allele Frequency">
    ##INFO=<ID=SB,Number=1,Type=Integer,Description="Phred-scaled strand bias at this position">
    ##INFO=<ID=DP4,Number=4,Type=Integer,Description="Counts for ref-forward bases, ref-reverse, alt-forward and alt-reverse bases">
    ##INFO=<ID=INDEL,Number=0,Type=Flag,Description="Indicates that the variant is an INDEL.">
    ##INFO=<ID=CONSVAR,Number=0,Type=Flag,Description="Indicates that the variant is a consensus variant (as opposed to a low frequency variant).">
    ##INFO=<ID=HRUN,Number=1,Type=Integer,Description="Homopolymer length to the right of report indel position">
    """
    dp = Column(Integer)
    dp4_ref_forward = Column(Integer)
    dp4_ref_reverse = Column(Integer)
    dp4_alt_forward = Column(Integer)
    dp4_alt_reverse = Column(Integer)
    vaf = Column(Float)
    strand_bias = Column(Integer)

    # fields replicated from Variant for performance reasons
    annotation = Column(String)
    annotation_impact = Column(String)
    biotype = Column(String)
    annotation_highest_impact = Column(String, index=True)
    gene_name = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)

    # fields replicated from sample for performance reasons
    date = Column(Date)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    ForeignKeyConstraint([sample, source], [Sample.id, Sample.source])
    ForeignKeyConstraint([variant_id], [Variant.variant_id])


class VariantCooccurrence(Base):

    __tablename__ = VARIANT_COOCCURRENCE_TABLE_NAME

    variant_id_one = Column(String, primary_key=True)
    variant_id_two = Column(String, primary_key=True)
    count = Column(Integer, default=0)

    ForeignKeyConstraint([variant_id_one], [Variant.variant_id])
    ForeignKeyConstraint([variant_id_two], [Variant.variant_id])


class CovigatorModule(enum.Enum):

    __constraint_name__ = COVIGATOR_MODULE_CONSTRAINT_NAME

    ACCESSOR = 1
    PROCESSOR = 2


class Log(Base):
    """
    The table that holds an ENA job
    """
    __tablename__ = LOG_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)

    start = Column(DateTime(timezone=True), nullable=False)
    end = Column(DateTime(timezone=True), nullable=False)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__), nullable=False)
    module = Column(Enum(CovigatorModule, name=CovigatorModule.__constraint_name__), nullable=False)

    has_error = Column(Boolean, default=False)
    error_message = Column(String)
    processed = Column(Integer)
    data = Column(JSON)


class Conservation(Base):
    """
    Table to hold conservation scores
    """
    __tablename__ = CONSERVATION_TABLE_NAME

    chromosome = Column(String, primary_key=True)
    start = Column(Integer, primary_key=True)
    end = Column(Integer, primary_key=True)
    conservation = Column(Float)
    conservation_sarbecovirus = Column(Float)
    conservation_vertebrates = Column(Float)


class PrecomputedVariantsPerSample(Base):

    __tablename__ = PRECOMPUTED_VARIANTS_PER_SAMPLE_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    count = Column(Integer)
    number_mutations = Column(Integer)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    gene_name = Column(String)


class PrecomputedSubstitutionsCounts(Base):

    __tablename__ = PRECOMPUTED_SUBSTITUTIONS_COUNTS_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    count = Column(Integer)
    reference = Column(String)
    alternate = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    gene_name = Column(String)


class PrecomputedIndelLength(Base):

    __tablename__ = PRECOMPUTED_INDEL_LENGTH_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    count = Column(Integer)
    length = Column(Integer)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    gene_name = Column(String)


class PrecomputedAnnotation(Base):

    __tablename__ = PRECOMPUTED_ANNOTATION_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    count = Column(Integer)
    annotation = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    gene_name = Column(String)


class PrecomputedOccurrence(Base):

    __tablename__ = PRECOMPUTED_OCCURRENCE_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    total = Column(Integer)
    count = Column(Integer)
    month = Column(String)
    frequency = Column(Float)
    frequency_by_month = Column(Float)
    variant_id = Column(String)
    hgvs_p = Column(String)
    gene_name = Column(String)
    annotation = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))


class PrecomputedDnDs(Base):

    __tablename__ = PRECOMPUTED_DN_DS_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    month = Column(Date)
    gene_name = Column(String)
    country = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    dn = Column(Integer)
    ds = Column(Integer)


class PrecomputedDnDsByDomain(Base):

    __tablename__ = PRECOMPUTED_DN_DS_BY_DOMAIN_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    month = Column(Date)
    domain = Column(String)
    country = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    dn = Column(Integer)
    ds = Column(Integer)


class PrecomputedTableCounts(Base):

    __tablename__ = PRECOMPUTED_TABLE_COUNTS_TABLE_NAME
    VIRTUAL_TABLE_COUNTRY = "Country"
    FACTOR_SOURCE = "source"

    id = Column(Integer, primary_key=True, autoincrement=True)
    table = Column(String)
    factor = Column(String)
    value = Column(String)
    count = Column(Integer)


class PrecomputedVariantAbundanceHistogram(Base):

    __tablename__ = PRECOMPUTED_VARIANT_ABUNDANCE_HIST_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    position_bin = Column(Integer)
    count_unique_variants = Column(Integer)
    count_variant_observations = Column(Integer)
    bin_size = Column(Integer)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))




import pandas as pd
from sqlalchemy.orm import Session
from covigator.configuration import Configuration
from covigator.dashboard.tabs.variants import BIN_SIZE_VALUES
from covigator.database.database import Database
from covigator.database.model import PrecomputedVariantsPerSample, PrecomputedSubstitutionsCounts, \
    VARIANT_OBSERVATION_TABLE_NAME, PrecomputedIndelLength, \
    PrecomputedAnnotation, VariantObservation, DataSource, PrecomputedOccurrence, PrecomputedDnDs, \
    SAMPLE_ENA_TABLE_NAME, SAMPLE_GISAID_TABLE_NAME, PrecomputedDnDsByDomain, PrecomputedTableCounts, Variant, \
    SubclonalVariantObservation, Sample, PrecomputedVariantAbundanceHistogram
from logzero import logger

from covigator.database.queries import Queries

NUMBER_TOP_OCCURRENCES = 1000


class Precomputer:

    def __init__(self, session: Session):
        self.session = session
        self.queries = Queries(session=self.session)

    def load_counts_variants_per_sample(self):

        # reads the data from the database
        sql_query = """
        select count(*) as count, counts.number_mutations, counts.source, counts.variant_type, counts.gene_name from (
            select count(*) as number_mutations, source, variant_type, gene_name, sample from {table_name}
            group by source, variant_type, gene_name, sample)
            as counts group by counts.number_mutations, counts.source, counts.variant_type, counts.gene_name;
            """.format(table_name=VARIANT_OBSERVATION_TABLE_NAME)
        data = pd.read_sql_query(sql_query, self.session.bind)
        sql_query = """
        select count(*) as count, counts.number_mutations, counts.source, counts.variant_type from (
            select count(*) as number_mutations, source, variant_type, sample from {table_name}
            group by source, variant_type, sample)
            as counts group by counts.number_mutations, counts.source, counts.variant_type;
            """.format(table_name=VARIANT_OBSERVATION_TABLE_NAME)
        data_without_gene = pd.read_sql_query(sql_query, self.session.bind)

        # delete all rows before starting
        self.session.query(PrecomputedVariantsPerSample).delete()
        self.session.commit()

        # stores the precomputed data
        database_rows = []
        for index, row in data.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedVariantsPerSample(
                count=row["count"],
                number_mutations=row["number_mutations"],
                source=row["source"],
                variant_type=row["variant_type"],
                gene_name=row["gene_name"] if row["gene_name"] is not None else "intergenic"
            ))

        for index, row in data_without_gene.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedVariantsPerSample(
                count=row["count"],
                number_mutations=row["number_mutations"],
                source=row["source"],
                variant_type=row["variant_type"]
            ))

        self.session.add_all(database_rows)
        self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedVariantsPerSample.__tablename__))

    def load_count_substitutions(self):

        sql_query = """
        select * from (select count(*) as count, reference, alternate, source, variant_type, gene_name 
            from {table_name}
            group by reference, alternate, source, variant_type, gene_name
            order by count desc) 
            as counts where counts.count > 10;
        """.format(table_name=VARIANT_OBSERVATION_TABLE_NAME)
        data = pd.read_sql_query(sql_query, self.session.bind)
        sql_query = """
        select * from (select count(*) as count, reference, alternate, source, variant_type 
            from {table_name}
            group by reference, alternate, source, variant_type
            order by count desc) 
            as counts where counts.count > 10;
        """.format(table_name=VARIANT_OBSERVATION_TABLE_NAME)
        data_without_gene = pd.read_sql_query(sql_query, self.session.bind)

        # delete all rows before starting
        self.session.query(PrecomputedSubstitutionsCounts).delete()
        self.session.commit()

        # stores the precomputed data
        database_rows = []
        for index, row in data.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedSubstitutionsCounts(
                count=row["count"],
                reference=row["reference"],
                alternate=row["alternate"],
                source=row["source"],
                variant_type=row["variant_type"],
                gene_name=row["gene_name"] if row["gene_name"] is not None else "intergenic"
            ))

        for index, row in data_without_gene.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedSubstitutionsCounts(
                count=row["count"],
                reference=row["reference"],
                alternate=row["alternate"],
                source=row["source"],
                variant_type=row["variant_type"]
            ))

        self.session.add_all(database_rows)
        self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedSubstitutionsCounts.__tablename__))

    def load_indel_length(self):

        sql_query = """
        select count(*) as count, length, source, gene_name 
            from {table_name}
            where variant_type != 'SNV'
            group by length, source, gene_name
            order by count asc;
        """.format(table_name=VARIANT_OBSERVATION_TABLE_NAME)
        data = pd.read_sql_query(sql_query, self.session.bind)
        sql_query = """
        select count(*) as count, length, source, gene_name 
            from {table_name}
            where variant_type != 'SNV'
            group by length, source, gene_name
            order by count desc;
        """.format(table_name=VARIANT_OBSERVATION_TABLE_NAME)
        data_without_gene = pd.read_sql_query(sql_query, self.session.bind)

        # delete all rows before starting
        self.session.query(PrecomputedIndelLength).delete()
        self.session.commit()

        # stores the precomputed data
        database_rows = []
        for index, row in data.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedIndelLength(
                count=row["count"],
                length=row["length"],
                source=row["source"],
                gene_name=row["gene_name"] if row["gene_name"] is not None else "intergenic"
            ))

        for index, row in data_without_gene.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedIndelLength(
                count=row["count"],
                length=row["length"],
                source=row["source"]
            ))

        self.session.add_all(database_rows)
        self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedIndelLength.__tablename__))

    def load_annotation(self):

        sql_query = """
        select count(*) as count, annotation_highest_impact, source, gene_name 
            from {table_name}
            where annotation_highest_impact is not null
            group by annotation_highest_impact, source, gene_name
            order by count asc;
        """.format(table_name=VARIANT_OBSERVATION_TABLE_NAME)
        data = pd.read_sql_query(sql_query, self.session.bind)
        sql_query = """
        select count(*) as count, annotation_highest_impact, source 
            from {table_name}
            where annotation_highest_impact is not null
            group by annotation_highest_impact, source
            order by count asc;
        """.format(table_name=VARIANT_OBSERVATION_TABLE_NAME)
        data_without_gene = pd.read_sql_query(sql_query, self.session.bind)

        # delete all rows before starting
        self.session.query(PrecomputedAnnotation).delete()
        self.session.commit()

        # stores the precomputed data
        database_rows = []
        for index, row in data.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedAnnotation(
                count=row["count"],
                annotation=row["annotation_highest_impact"],
                source=row["source"],
                gene_name=row["gene_name"] if row["gene_name"] is not None else "intergenic"
            ))

        for index, row in data_without_gene.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedAnnotation(
                count=row["count"],
                annotation=row["annotation_highest_impact"],
                source=row["source"]
            ))

        self.session.add_all(database_rows)
        self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedAnnotation.__tablename__))

    def load_top_occurrences(self):

        # gets the top occurrent variants for each source and overall
        top_occurring_variants_ena = None
        try:
            top_occurring_variants_ena = self.queries.get_top_occurring_variants(
                top=NUMBER_TOP_OCCURRENCES, source=DataSource.ENA.name)
        except ValueError:
            logger.error("No top occurrences for ENA data")

        top_occurring_variants_gisaid = None
        try:
            top_occurring_variants_gisaid = self.queries.get_top_occurring_variants(
                top=NUMBER_TOP_OCCURRENCES, source=DataSource.GISAID.name)
        except ValueError:
            logger.error("No top occurrences for GISAID data")

        top_occurring_variants = None
        try:
            top_occurring_variants = self.queries.get_top_occurring_variants(top=NUMBER_TOP_OCCURRENCES)
        except ValueError:
            logger.error("No top occurrences")

        # delete all rows before starting
        self.session.query(PrecomputedOccurrence).delete()
        self.session.commit()

        database_rows = []
        # stores the precomputed data
        if top_occurring_variants_ena is not None:
            for index, row in top_occurring_variants_ena.iterrows():
                # add entries per gene
                database_rows.append(self._row_to_top_occurrence(row, source=DataSource.ENA))

        if top_occurring_variants_gisaid is not None:
            for index, row in top_occurring_variants_gisaid.iterrows():
                # add entries per gene
                database_rows.append(self._row_to_top_occurrence(row, source=DataSource.GISAID))

        if top_occurring_variants is not None:
            for index, row in top_occurring_variants.iterrows():
                # add entries per gene
                database_rows.append(self._row_to_top_occurrence(row))

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedOccurrence.__tablename__))

    def _row_to_top_occurrence(self, row, source=None):
        return PrecomputedOccurrence(
            total=row["total"],
            frequency=row["frequency"],
            variant_id=row["variant_id"],
            hgvs_p=row["hgvs_p"],
            gene_name=row["gene_name"],
            annotation=row["annotation_highest_impact"],
            source=source,
            month=row["month"],
            count=row["count"],
            frequency_by_month=row["frequency_by_month"],
        )

    def load_dn_ds(self):
        sql_query_ds_ena = """
                select count(*) as ds, date_trunc('month', s.collection_date) as month, vo.gene_name, s.country 
                from {variant_observation_table} as vo join {sample_ena_table} as s on vo.sample = s.run_accession 
                where vo.source = 'ENA' and vo.annotation_highest_impact = 'synonymous_variant' 
                    and s.collection_date is not null
                group by date_trunc('month', s.collection_date), vo.gene_name, s.country;
                """.format(variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
                           sample_ena_table=SAMPLE_ENA_TABLE_NAME)
        data_ds_ena = pd.read_sql_query(sql_query_ds_ena, self.session.bind)

        sql_query_dn_ena = """
                select count(*) as dn, date_trunc('month', s.collection_date) as month, vo.gene_name, s.country 
                from {variant_observation_table} as vo join {sample_ena_table} as s on vo.sample = s.run_accession 
                where vo.source = 'ENA' and vo.annotation_highest_impact = 'missense_variant' 
                    and s.collection_date is not null
                group by date_trunc('month', s.collection_date), vo.gene_name, s.country;
                """.format(variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
                           sample_ena_table=SAMPLE_ENA_TABLE_NAME)
        data_dn_ena = pd.read_sql_query(sql_query_dn_ena, self.session.bind)

        data_ena = pd.merge(left=data_ds_ena, right=data_dn_ena, on=["month", "gene_name", "country"], how='outer').fillna(0)

        sql_query_ds_gisaid = """
                select count(*) as ds, date_trunc('month', s.date) as month, vo.gene_name, s.country 
                from {variant_observation_table} as vo join {sample_gisaid_table} as s on vo.sample = s.run_accession 
                    and s.date is not null
                where vo.source = 'GISAID' and vo.annotation_highest_impact = 'synonymous_variant' 
                group by date_trunc('month', s.date), vo.gene_name, s.country;
                """.format(variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
                           sample_gisaid_table=SAMPLE_GISAID_TABLE_NAME)
        data_ds_gisaid = pd.read_sql_query(sql_query_ds_gisaid, self.session.bind)

        sql_query_dn_gisaid = """
                select count(*) as dn, date_trunc('month', s.date) as month, vo.gene_name, s.country 
                from {variant_observation_table} as vo join {sample_gisaid_table} as s on vo.sample = s.run_accession 
                where vo.source = 'GISAID' and vo.annotation_highest_impact = 'missense_variant'
                    and s.date is not null 
                group by date_trunc('month', s.date), vo.gene_name, s.country;
                """.format(variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
                           sample_gisaid_table=SAMPLE_GISAID_TABLE_NAME)
        data_dn_gisaid = pd.read_sql_query(sql_query_dn_gisaid, self.session.bind)

        data_gisaid = pd.merge(left=data_ds_gisaid, right=data_dn_gisaid, on=["month", "gene_name", "country"],
                            how='outer').fillna(0)

        # delete all rows before starting
        self.session.query(PrecomputedDnDs).delete()
        self.session.commit()

        database_rows = []
        if data_ena is not None:
            for index, row in data_ena.iterrows():
                # add entries per gene
                database_rows.append(PrecomputedDnDs(
                    month=row["month"],
                    gene_name=row["gene_name"],
                    country=row["country"],
                    dn=row["dn"],
                    ds=row["ds"],
                    source=DataSource.ENA
                ))
        if data_gisaid is not None:
            for index, row in data_gisaid.iterrows():
                # add entries per gene
                database_rows.append(PrecomputedDnDs(
                    month=row["month"],
                    gene_name=row["gene_name"],
                    country=row["country"],
                    dn=row["dn"],
                    ds=row["ds"],
                    source=DataSource.GISAID
                ))

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedDnDs.__tablename__))


    def load_dn_ds_by_domain(self):
        sql_query_ds_ena = """
                select count(*) as ds, date_trunc('month', s.collection_date) as month, vo.pfam_name, s.country 
                from {variant_observation_table} as vo join {sample_ena_table} as s on vo.sample = s.run_accession 
                where vo.source = 'ENA' and vo.annotation_highest_impact = 'synonymous_variant' 
                    and vo.pfam_name is not null and s.collection_date is not null
                group by date_trunc('month', s.collection_date), vo.pfam_name, s.country;
                """.format(variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
                           sample_ena_table=SAMPLE_ENA_TABLE_NAME)
        data_ds_ena = pd.read_sql_query(sql_query_ds_ena, self.session.bind)

        sql_query_dn_ena = """
                select count(*) as dn, date_trunc('month', s.collection_date) as month, vo.pfam_name, s.country 
                from {variant_observation_table} as vo join {sample_ena_table} as s on vo.sample = s.run_accession 
                where vo.source = 'ENA' and vo.annotation_highest_impact = 'missense_variant' 
                    and vo.pfam_name is not null and s.collection_date is not null
                group by date_trunc('month', s.collection_date), vo.pfam_name, s.country;
                """.format(variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
                           sample_ena_table=SAMPLE_ENA_TABLE_NAME)
        data_dn_ena = pd.read_sql_query(sql_query_dn_ena, self.session.bind)

        data_ena = pd.merge(left=data_ds_ena, right=data_dn_ena, on=["month", "pfam_name", "country"], how='outer').fillna(0)

        sql_query_ds_gisaid = """
                select count(*) as ds, date_trunc('month', s.date) as month, vo.pfam_name, s.country 
                from {variant_observation_table} as vo join {sample_gisaid_table} as s on vo.sample = s.run_accession 
                where vo.source = 'GISAID' and vo.annotation_highest_impact = 'synonymous_variant' 
                    and vo.pfam_name is not null and s.date is not null
                group by date_trunc('month', s.date), vo.pfam_name, s.country;
                """.format(variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
                           sample_gisaid_table=SAMPLE_GISAID_TABLE_NAME)
        data_ds_gisaid = pd.read_sql_query(sql_query_ds_gisaid, self.session.bind)

        sql_query_dn_gisaid = """
                select count(*) as dn, date_trunc('month', s.date) as month, vo.pfam_name, s.country 
                from {variant_observation_table} as vo join {sample_gisaid_table} as s on vo.sample = s.run_accession 
                where vo.source = 'GISAID' and vo.annotation_highest_impact = 'missense_variant' 
                    and vo.pfam_name is not null and s.date is not null
                group by date_trunc('month', s.date), vo.pfam_name, s.country;
                """.format(variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
                           sample_gisaid_table=SAMPLE_GISAID_TABLE_NAME)
        data_dn_gisaid = pd.read_sql_query(sql_query_dn_gisaid, self.session.bind)

        data_gisaid = pd.merge(left=data_ds_gisaid, right=data_dn_gisaid, on=["month", "pfam_name", "country"],
                            how='outer').fillna(0)

        # delete all rows before starting
        self.session.query(PrecomputedDnDs).delete()
        self.session.commit()

        database_rows = []
        if data_ena is not None:
            for index, row in data_ena.iterrows():
                # add entries per gene
                database_rows.append(PrecomputedDnDsByDomain(
                    month=row["month"],
                    domain=row["pfam_name"],
                    country=row["country"],
                    dn=row["dn"],
                    ds=row["ds"],
                    source=DataSource.ENA
                ))
        if data_gisaid is not None:
            for index, row in data_gisaid.iterrows():
                # add entries per gene
                database_rows.append(PrecomputedDnDsByDomain(
                    month=row["month"],
                    domain=row["pfam_name"],
                    country=row["country"],
                    dn=row["dn"],
                    ds=row["ds"],
                    source=DataSource.GISAID
                ))

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedDnDsByDomain.__tablename__))

    def load_table_counts(self):

        count_variants = self.queries.count_variants(cache=False)
        count_samples = self.queries.count_samples(cache=False)
        count_samples_ena = self.queries.count_samples(source=DataSource.ENA.name, cache=False)
        count_samples_gisaid = self.queries.count_samples(source=DataSource.GISAID.name, cache=False)
        count_variant_observations = self.queries.count_variant_observations(cache=False)
        count_variant_observations_ena = self.queries.count_variant_observations(source=DataSource.ENA.name, cache=False)
        count_variant_observations_gisaid = self.queries.count_variant_observations(source=DataSource.GISAID.name, cache=False)
        count_subclonal_variant_observations = self.queries.count_subclonal_variant_observations(cache=False)
        count_countries = self.queries.count_countries(cache=False)
        count_countries_ena = self.queries.count_countries(source=DataSource.ENA.name, cache=False)
        count_countries_gisaid = self.queries.count_countries(source=DataSource.GISAID.name, cache=False)

        # delete all rows before starting
        self.session.query(PrecomputedTableCounts).delete()
        self.session.commit()

        database_rows = [
            PrecomputedTableCounts(table=Variant.__name__, count=count_variants),
            PrecomputedTableCounts(table=VariantObservation.__name__, count=count_variant_observations),
            PrecomputedTableCounts(
                table=VariantObservation.__name__, count=count_variant_observations_ena,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.ENA.name),
            PrecomputedTableCounts(
                table=VariantObservation.__name__, count=count_variant_observations_gisaid,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.GISAID.name),
            PrecomputedTableCounts(
                table=SubclonalVariantObservation.__name__, count=count_subclonal_variant_observations),
            PrecomputedTableCounts(table=Sample.__name__, count=count_samples),
            PrecomputedTableCounts(table=Sample.__name__, count=count_samples_ena,
                                   factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.ENA.name),
            PrecomputedTableCounts(table=Sample.__name__, count=count_samples_gisaid,
                                   factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.GISAID.name),
            PrecomputedTableCounts(table=PrecomputedTableCounts.VIRTUAL_TABLE_COUNTRY, count=count_countries),
            PrecomputedTableCounts(table=PrecomputedTableCounts.VIRTUAL_TABLE_COUNTRY, count=count_countries_ena,
                                   factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.ENA.name),
            PrecomputedTableCounts(table=PrecomputedTableCounts.VIRTUAL_TABLE_COUNTRY, count=count_countries_gisaid,
                                   factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.GISAID.name),
        ]

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedTableCounts.__tablename__))

    def load_variant_abundance_histogram(self):

        histograms = []
        for bin_size in BIN_SIZE_VALUES:
            histogram = self.queries.get_variant_abundance_histogram(bin_size=bin_size, cache=False)
            histogram["bin_size"] = bin_size
            histogram["source"] = None
            histograms.append(histogram)

        for bin_size in BIN_SIZE_VALUES:
            histogram = self.queries.get_variant_abundance_histogram(
                bin_size=bin_size, source=DataSource.ENA.name, cache=False)
            histogram["bin_size"] = bin_size
            histogram["source"] = DataSource.ENA
            histograms.append(histogram)

        for bin_size in BIN_SIZE_VALUES:
            histogram = self.queries.get_variant_abundance_histogram(
                bin_size=bin_size, source=DataSource.GISAID.name, cache=False)
            histogram["bin_size"] = bin_size
            histogram["source"] = DataSource.GISAID
            histograms.append(histogram)

        # delete all rows before starting
        self.session.query(PrecomputedVariantAbundanceHistogram).delete()
        self.session.commit()

        database_rows = []
        if len(histograms) > 0:
            for index, row in pd.concat(histograms).iterrows():
                # add entries per gene
                database_rows.append(PrecomputedVariantAbundanceHistogram(
                    position_bin=row["position_bin"],
                    count_unique_variants=row["count_unique_variants"],
                    count_variant_observations=row["count_variant_observations"],
                    bin_size=row["bin_size"],
                    source=row["source"],
                ))

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(
            len(database_rows), PrecomputedVariantAbundanceHistogram.__tablename__))


if __name__ == '__main__':
    database = Database(initialize=True, config=Configuration())
    precomputer = Precomputer(session=database.get_database_session())
    #precomputer.load_dn_ds_by_domain()
    #precomputer.load_dn_ds()
    #precomputer.load_table_counts()
    precomputer.load_variant_abundance_histogram()



from unittest import TestCase
from covigator.misc.compression import compress_sequence, decompress_sequence


class TestCompression(TestCase):

    def test_compression(self):
        test_sequence = "ACGTACGTACGT"
        compressed_sequence = compress_sequence(test_sequence)
        decompressed_sequence = decompress_sequence(compressed_sequence)
        self.assertEqual(test_sequence, decompressed_sequence)


import abc
import traceback
from datetime import datetime
from typing import Callable
import typing as typing
from dask.distributed import Client
from dask.distributed import wait
from sqlalchemy.orm import Session
from logzero import logger
import covigator
import covigator.configuration
from covigator.configuration import Configuration
from covigator.database.database import Database, session_scope
from covigator.database.model import Log, DataSource, CovigatorModule, JobStatus, JobEna, JobGisaid
from covigator.database.queries import Queries


class AbstractProcessor:

    def __init__(self, database: Database, dask_client: Client, data_source: DataSource, config: Configuration):
        self.data_source = data_source
        self.config = config
        self.start_time = datetime.now()
        self.has_error = False
        self.error_message = None
        self.database = database
        assert self.database is not None, "Empty database"
        self.dask_client = dask_client
        assert self.dask_client is not None, "Empty dask client"

    def process(self):
        logger.info("Starting processor")
        session = self.database.get_database_session()
        queries = Queries(session)
        count = 0
        try:
            futures = []
            while True:
                job = queries.find_first_pending_job(self.data_source)
                if not job:
                    logger.info("No more jobs to process after sending {} runs to process".format(count))
                    break

                # it has to update the status before doing anything so this processor does not read it again
                job.status = JobStatus.QUEUED
                job.queued_at = datetime.now()
                session.commit()

                # sends the run for processing
                futures.append(self._process_run(run_accession=job.run_accession))
                count += 1

                if len(futures) > self.config.batch_size:
                    # waits for a batch to finish before sending more
                    self.dask_client.gather(futures=futures)
                    futures = []

            # waits for all to finish
            if len(futures) > 0:
                wait(futures)
            logger.info("Processor finished!")
        except Exception as e:
            logger.exception(e)
            session.rollback()
            self.error_message = self._get_traceback_from_exception(e)
            self.has_error = True
        finally:
            self._write_execution_log(session, count, data_source=self.data_source)
            self.dask_client.shutdown()
            session.close()
            logger.info("Finished processor")

    @staticmethod
    def run_job(config: Configuration, run_accession: str, start_status: JobStatus, end_status: JobStatus,
                error_status: JobStatus, data_source: DataSource,
                function: Callable[[typing.Union[JobEna, JobGisaid], Queries, Configuration], None]) -> str or None:
        """
        Runs a function on a job, if anything goes wrong or does not fit in the DB it returns None in order to
        stop the execution of subsequent jobs.
        """
        covigator.configuration.initialise_logs(config.logfile_processor, sample_id=run_accession)
        if run_accession is not None:
            try:
                with session_scope(config=config) as session:
                    queries = Queries(session)
                    job = queries.find_job_by_accession_and_status(
                        run_accession=run_accession, status=start_status, data_source=data_source)
                    if job is not None:
                        function(job, queries, config)
                        if end_status is not None:
                            job.status = end_status
                            if job.status == JobStatus.FINISHED:
                                sample = queries.find_sample_by_accession(
                                    run_accession=run_accession, source=data_source)
                                sample.finished = True
                    else:
                        logger.warning("Expected ENA job {} in status {}".format(run_accession, start_status))
                        run_accession = None
            except Exception as e:
                # captures any possible exception happening, but logs it in the DB
                if error_status is not None:
                    AbstractProcessor._log_error_in_job(
                        config=config, run_accession=run_accession, exception=e, status=error_status,
                        data_source=data_source)
                    run_accession = None
                else:
                    logger.warning("Error processing a job that does not stop the workflow!")
        return run_accession

    @abc.abstractmethod
    def _process_run(self, run_accession: str):
        pass

    def _write_execution_log(self, session: Session, count, data_source: DataSource):
        end_time = datetime.now()
        session.add(Log(
            start=self.start_time,
            end=end_time,
            source=data_source,
            module=CovigatorModule.PROCESSOR,
            has_error=self.has_error,
            processed=count,
            data={
                "processed": count
            }
        ))
        session.commit()

    @staticmethod
    def _get_traceback_from_exception(e):
        return "".join(traceback.format_exception(
            etype=type(e), value=e, tb=e.__traceback__))

    @staticmethod
    def _log_error_in_job(config: Configuration, run_accession: str, exception: Exception, status: JobStatus, data_source: DataSource):
        with session_scope(config=config) as session:
            logger.exception(exception)
            logger.info("Error on job {} on state {}: {}".format(run_accession, status, str(exception)))
            job = Queries(session).find_job_by_accession(run_accession=run_accession, data_source=data_source)
            job.status = status
            job.failed_at = datetime.now()
            job.error_message = AbstractProcessor._get_traceback_from_exception(exception)


import functools
from datetime import date, datetime
from typing import List, Union
import numpy as np
import pandas as pd
from scipy.spatial.distance import squareform
from logzero import logger
from skbio.stats.distance import DissimilarityMatrix
from sklearn.cluster import OPTICS
from sqlalchemy import and_, desc, asc, func, String, DateTime, cast
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.orm import Session, aliased
from sqlalchemy.sql.sqltypes import NullType

from covigator.database.model import Log, DataSource, CovigatorModule, SampleEna, JobEna, JobStatus, VariantObservation, \
    Gene, Variant, VariantCooccurrence, Conservation, JobGisaid, SampleGisaid, SubclonalVariantObservation, \
    PrecomputedVariantsPerSample, PrecomputedSubstitutionsCounts, PrecomputedIndelLength, VariantType, \
    PrecomputedAnnotation, PrecomputedOccurrence, PrecomputedTableCounts, Sample, PrecomputedVariantAbundanceHistogram, \
    VARIANT_OBSERVATION_TABLE_NAME
from covigator.exceptions import CovigatorQueryException, CovigatorDashboardMissingPrecomputedData

SYNONYMOUS_VARIANT = "synonymous_variant"


class Queries:

    def __init__(self, session: Session):
        self.session = session

    def find_job_by_accession_and_status(
            self, run_accession: str, status: JobStatus, data_source: DataSource) -> Union[JobEna, JobGisaid]:
        if data_source == DataSource.ENA:
            return self.session.query(JobEna) \
                .filter(and_(JobEna.run_accession == run_accession, JobEna.status == status)) \
                .first()
        elif data_source == DataSource.GISAID:
            return self.session.query(JobGisaid)\
                .filter(and_(JobGisaid.run_accession == run_accession, JobGisaid.status == status)) \
                .first()
        else:
            raise ValueError("Bad data source {}".format(data_source))

    def find_job_by_accession(self, run_accession: str, data_source: DataSource) -> Union[JobEna, JobGisaid]:
        if data_source == DataSource.ENA:
            return self.session.query(JobEna).filter(JobEna.run_accession == run_accession).first()
        elif data_source == DataSource.GISAID:
            return self.session.query(JobGisaid).filter(JobGisaid.run_accession == run_accession).first()
        else:
            raise ValueError("Bad data source {}".format(data_source))

    def find_first_pending_job(self, data_source: DataSource) -> Union[JobEna, JobGisaid]:
        if data_source == DataSource.ENA:
            return self.session.query(JobEna) \
                .filter(JobEna.status == JobStatus.PENDING) \
                .order_by(JobEna.created_at.desc()) \
                .first()
        elif data_source == DataSource.GISAID:
            return self.session.query(JobGisaid) \
                .filter(JobGisaid.status == JobStatus.PENDING) \
                .order_by(JobGisaid.created_at.desc()) \
                .first()
        else:
            raise ValueError("Bad data source {}".format(data_source))

    def count_jobs_in_queue(self, data_source):
        if data_source == DataSource.ENA:
            count = self.session.query(JobEna).filter(JobEna.status.in_((
                JobStatus.QUEUED, JobStatus.DOWNLOADED, JobStatus.PROCESSED, JobStatus.LOADED))).count()
        elif data_source == DataSource.GISAID:
            count = self.session.query(JobGisaid).filter(JobGisaid.status.in_((
                JobStatus.QUEUED, JobStatus.DOWNLOADED, JobStatus.PROCESSED, JobStatus.LOADED))).count()
        else:
            raise ValueError("Bad data source {}".format(data_source))
        return count

    def find_sample_by_accession(self, run_accession: str, source: DataSource) -> Union[SampleEna, SampleGisaid]:
        if source == DataSource.ENA:
            sample = self.session.query(SampleEna).filter(SampleEna.run_accession == run_accession).first()
        elif source == DataSource.GISAID:
            sample = self.session.query(SampleGisaid).filter(SampleGisaid.run_accession == run_accession).first()
        else:
            raise CovigatorQueryException("Bad query trying to fetch a sample")
        return sample

    def get_countries(self, source: str = None) -> List[str]:
        countries = []
        if source == DataSource.ENA.name or source is None:
            countries = countries + [c for c, in self.session.query(SampleEna.country).filter(
                SampleEna.finished).distinct().all()]
        if source == DataSource.GISAID.name or source is None:
            countries = countries + [c for c, in self.session.query(SampleGisaid.country).filter(
                SampleGisaid.finished).distinct().all()]
        return sorted(list(set(countries)))

    def get_variants_per_sample(self, data_source: str, genes: List[str], variant_types: List[str]):
        """
        Returns a DataFrame with columns: number_mutations, count, type
        where type: SNV, insertion or deletion
        """
        query = self.session.query(PrecomputedVariantsPerSample)
        if data_source is not None:
            query = query.filter(PrecomputedVariantsPerSample.source == data_source)
        if genes is not None and genes:
            query = query.filter(PrecomputedVariantsPerSample.gene_name.in_(genes))
        else:
            query = query.filter(PrecomputedVariantsPerSample.gene_name == None)
        if variant_types is not None and variant_types:
            query = query.filter(PrecomputedVariantsPerSample.variant_type.in_(variant_types))

        data = pd.read_sql(query.statement, self.session.bind)
        if data.shape[0] > 0:
            data.variant_type = data.variant_type.transform(lambda x: x.name if x else x)
        return data[["number_mutations", "variant_type", "count"]] \
            .groupby(["number_mutations", "variant_type"]).sum().reset_index()

    def get_indel_lengths(self, data_source, genes):
        query = self.session.query(PrecomputedIndelLength)
        if data_source is not None:
            query = query.filter(PrecomputedIndelLength.source == data_source)
        if genes is not None and genes:
            query = query.filter(PrecomputedIndelLength.gene_name.in_(genes))
        else:
            query = query.filter(PrecomputedIndelLength.gene_name == None)

        data = pd.read_sql(query.statement, self.session.bind)
        if data.shape[0] > 0:
            data["tmp_variant_type"] = data["length"].transform(
                lambda x: VariantType.INSERTION.name if x > 0 else VariantType.DELETION.name)
            data["inframe"] = data["length"].transform(lambda x: "INFRAME" if x % 3 == 0 else "FRAMESHIFT")
            data["variant_type"] = data[["tmp_variant_type", "inframe"]].apply(
                lambda x: "{}_{}".format(x[0], x[1]), axis=1)
            return data[["length", "variant_type", "count"]] \
                .groupby(["length", "variant_type", ]).sum().reset_index() \
                .sort_values("length", ascending=True)
        return data

    def get_annotations(self, data_source, genes):
        query = self.session.query(PrecomputedAnnotation)
        if data_source is not None:
            query = query.filter(PrecomputedAnnotation.source == data_source)
        if genes is not None and genes:
            query = query.filter(PrecomputedAnnotation.gene_name.in_(genes))
        else:
            query = query.filter(PrecomputedAnnotation.gene_name == None)

        data = pd.read_sql(query.statement, self.session.bind)
        if data.shape[0] > 0:
            return data[["annotation", "count"]] \
                .groupby(["annotation", ]).sum().reset_index() \
                .sort_values("count", ascending=False)
        return data

    def get_substitutions(self, data_source, genes, variant_types):
        query = self.session.query(PrecomputedSubstitutionsCounts)
        if variant_types is not None and variant_types:
            query = query.filter(PrecomputedSubstitutionsCounts.variant_type.in_(variant_types))
        if data_source is not None:
            query = query.filter(PrecomputedSubstitutionsCounts.source == data_source)
        if genes is not None and genes:
            query = query.filter(PrecomputedSubstitutionsCounts.gene_name.in_(genes))
        else:
            query = query.filter(PrecomputedSubstitutionsCounts.gene_name == None)

        data = pd.read_sql(query.statement, self.session.bind)
        if data.shape[0] > 0:
            data.variant_type = data.variant_type.transform(lambda x: x.name if x else x)
            data["substitution"] = data[["reference", "alternate"]].apply(lambda x: "{}>{}".format(x[0], x[1]), axis=1)
            total_number_variants = data["count"].sum()
            data = data[["substitution", "variant_type", "count"]] \
                .groupby(["substitution", "variant_type"]).sum().reset_index()\
                .sort_values("count", ascending=False)
            data["rate"] = (data["count"] / total_number_variants).transform(lambda x: "{} %".format(round(x * 100, 1)))
            data = data.head(12)
        return data

    @functools.lru_cache()
    def get_accumulated_samples_by_country(
            self, data_source: DataSource, countries: List[str], min_samples=100) -> pd.DataFrame:
        """
        Returns a DataFrame with columns: data, country, cumsum, count
        """
        samples_ena = None

        if data_source is None or data_source == DataSource.ENA.name:
            query = self.session.query(
                func.count().label("count"), SampleEna.collection_date.label("date"), SampleEna.country) \
                .filter(SampleEna.finished) \
                .group_by(SampleEna.collection_date, SampleEna.country)
            if countries:
                query = query.filter(SampleEna.country.in_(countries))
            samples_ena = pd.read_sql(query.statement, self.session.bind).astype(
                {'date': 'datetime64', 'count': 'float64'})

        samples_gisaid = None
        if data_source is None or data_source == DataSource.GISAID.name:
            query = self.session.query(
                func.count().label("count"), SampleGisaid.date, SampleGisaid.country) \
                .filter(SampleGisaid.finished)\
                .group_by(SampleGisaid.date, SampleGisaid.country)
            if countries:
                query = query.filter(SampleGisaid.country.in_(countries))
            samples_gisaid = pd.read_sql(query.statement, self.session.bind).astype(
                {'date': 'datetime64', 'count': 'float64'})

        if samples_gisaid is None:
            samples = samples_ena
        elif samples_ena is None:
            samples = samples_gisaid
        else:
            samples = pd.concat([samples_gisaid, samples_ena]).groupby(['date', 'country']).sum().reset_index()

        filled_table = None
        if samples is not None and samples.shape[0] > 0:
            # accumulates count ordered by date
            samples['cumsum'] = samples.groupby(['country'])['count'].cumsum()

            # creates empty table with all pairwise combinations of date and country
            dates = samples.date[~samples.date.isna()].sort_values().unique()
            countries = samples[(~samples.country.isna()) & (samples["cumsum"] >= min_samples)] \
                .sort_values("cumsum", ascending=False).country.unique()

            empty_table = pd.DataFrame(
                index=pd.MultiIndex.from_product([dates, countries], names=["date", "country"]))
            empty_table["count"] = 0

            # adds values into empty table
            filled_table = pd.merge(
                left=empty_table.reset_index(),
                right=samples.loc[:, ["date", "country", "count"]],
                on=["date", "country"],
                how='left',
                suffixes=("_x", "_y")).fillna(0)

            filled_table["count"] = filled_table.count_x + filled_table.count_y
            filled_table['cumsum'] = filled_table.groupby(['country'])['count'].cumsum()

        return filled_table

    def get_sample_months(self, pattern) -> List[datetime]:
        dates_ena = [d.strftime(pattern) for d, in
                     self.session.query(SampleEna.collection_date).filter(
                         and_(SampleEna.finished, SampleEna.collection_date.isnot(None))).distinct().all()]
        dates_gisaid = [d.strftime(pattern) for d, in
                     self.session.query(SampleGisaid.date).filter(
                         and_(SampleGisaid.finished, SampleGisaid.date.isnot(None))).distinct().all()]
        return sorted(set(dates_ena + dates_gisaid))

    def get_gene(self, gene_name: str):
        return self.session.query(Gene).filter(Gene.name == gene_name).first()

    def get_genes(self):
        return self.session.query(Gene).order_by(Gene.start).all()

    def get_non_synonymous_variants_by_region(self, start, end, source) -> pd.DataFrame:
        query = self.session.query(VariantObservation.position,
                                      VariantObservation.annotation,
                                      VariantObservation.hgvs_p,
                                      func.count().label("count_occurrences"))\
            .filter(and_(VariantObservation.position >= start, VariantObservation.position <= end,
                         VariantObservation.annotation != SYNONYMOUS_VARIANT))
        if source == DataSource.ENA.name:
            query = query.filter(VariantObservation.source == DataSource.ENA)
        elif source == DataSource.GISAID.name:
            query = query.filter(VariantObservation.source == DataSource.GISAID)
        subquery = query.group_by(VariantObservation.position, VariantObservation.annotation, VariantObservation.hgvs_p).subquery()
        return pd.read_sql(
            self.session.query(subquery).filter(subquery.c.count_occurrences > 1).statement, self.session.bind)

    def get_variants_by_sample(self, sample_id) -> List[VariantObservation]:
        return self.session.query(VariantObservation) \
            .filter(VariantObservation.sample == sample_id) \
            .order_by(VariantObservation.position, VariantObservation.reference, VariantObservation.alternate) \
            .all()

    def get_variant_cooccurrence(self, variant_one: Variant, variant_two: Variant) -> VariantCooccurrence:
        return self.session.query(VariantCooccurrence) \
            .filter(and_(VariantCooccurrence.variant_id_one == variant_one.variant_id,
                         VariantCooccurrence.variant_id_two == variant_two.variant_id)) \
            .first()

    @functools.lru_cache()
    def count_samples(self, source: str = None, cache=True) -> int:
        if cache:
            query = self.session.query(PrecomputedTableCounts.count)
            if source is not None:
                query = query.filter(and_(
                    PrecomputedTableCounts.table == Sample.__name__,
                    PrecomputedTableCounts.factor == PrecomputedTableCounts.FACTOR_SOURCE,
                    PrecomputedTableCounts.value == source
                ))
            else:
                query = query.filter(and_(
                    PrecomputedTableCounts.table == Sample.__name__,
                    PrecomputedTableCounts.factor == None
                ))
            result = query.first()
            if result is None:
                raise CovigatorDashboardMissingPrecomputedData
            count = result.count
        else:
            count = 0
            if source is None or source == DataSource.ENA.name:
                count += self.session.query(SampleEna).filter(SampleEna.finished).count()
            if source is None or source == DataSource.GISAID.name:
                count += self.session.query(SampleGisaid).filter(SampleGisaid.finished).count()
        return count

    @functools.lru_cache()
    def count_countries(self, source: str = None, cache=True):
        if cache:
            query = self.session.query(PrecomputedTableCounts.count)
            if source is not None:
                query = query.filter(and_(
                    PrecomputedTableCounts.table == PrecomputedTableCounts.VIRTUAL_TABLE_COUNTRY,
                    PrecomputedTableCounts.factor == PrecomputedTableCounts.FACTOR_SOURCE,
                    PrecomputedTableCounts.value == source
                ))
            else:
                query = query.filter(and_(
                    PrecomputedTableCounts.table == PrecomputedTableCounts.VIRTUAL_TABLE_COUNTRY,
                    PrecomputedTableCounts.factor == None
                ))
            result = query.first()
            if result is None:
                raise CovigatorDashboardMissingPrecomputedData
            count = result.count
        else:
            count = len(self.get_countries(source=source))
        return count

    @functools.lru_cache()
    def count_variants(self, cache=True):
        if cache:
            result = self.session.query(PrecomputedTableCounts.count) \
                .filter(PrecomputedTableCounts.table == Variant.__name__).first()
            if result is None:
                raise CovigatorDashboardMissingPrecomputedData
            count = result.count
        else:
            count = self.session.query(Variant).count()
        return count

    @functools.lru_cache()
    def count_insertions(self):
        return self.session.query(Variant).filter(func.length(Variant.alternate) > 1).count()

    @functools.lru_cache()
    def count_deletions(self):
        return self.session.query(Variant).filter(func.length(Variant.reference) > 1).count()

    @functools.lru_cache()
    def count_variant_observations(self, source: str = None, cache=True):
        if cache:
            query = self.session.query(PrecomputedTableCounts.count)
            if source is not None:
                query = query.filter(and_(
                    PrecomputedTableCounts.table == VariantObservation.__name__,
                    PrecomputedTableCounts.factor == PrecomputedTableCounts.FACTOR_SOURCE,
                    PrecomputedTableCounts.value == source
                ))
            else:
                query = query.filter(and_(
                    PrecomputedTableCounts.table == VariantObservation.__name__,
                    PrecomputedTableCounts.factor == None
                ))
            result = query.first()
            if result is None:
                raise CovigatorDashboardMissingPrecomputedData
            count = result.count
        else:
            query = self.session.query(VariantObservation)
            if source == DataSource.GISAID.name or source == DataSource.ENA.name:
                query = query.filter(VariantObservation.source == source)
            count = query.count()
        return count

    @functools.lru_cache()
    def count_subclonal_variant_observations(self, cache=True):
        if cache:
            query = self.session.query(PrecomputedTableCounts.count) \
                .filter(PrecomputedTableCounts.table == SubclonalVariantObservation.__name__)
            result = query.first()
            count = result.count
        else:
            count = self.session.query(SubclonalVariantObservation).count()
        return count

    @functools.lru_cache()
    def get_date_of_first_sample(self, source: DataSource = DataSource.ENA) -> date:
        """
        Returns the date of the earliest ENA sample loaded in the database
        """
        if source == DataSource.ENA:
            result = self.session.query(SampleEna.collection_date).filter(
                and_(SampleEna.finished, SampleEna.collection_date.isnot(None))) \
                .order_by(asc(SampleEna.collection_date)).first()
        elif source == DataSource.GISAID:
            result = self.session.query(SampleGisaid.date).filter(
                and_(SampleGisaid.finished, SampleGisaid.date.isnot(None))) \
                .order_by(asc(SampleGisaid.date)).first()
        else:
            raise CovigatorQueryException("No valid data source for query of first sample")
        return result[0] if result is not None else result

    @functools.lru_cache()
    def get_date_of_most_recent_sample(self, source: DataSource = DataSource.ENA) -> date:
        """
        Returns the date of the latest ENA sample loaded in the database
        """
        if source == DataSource.ENA:
            result = self.session.query(SampleEna.collection_date).filter(
                and_(SampleEna.finished, SampleEna.collection_date.isnot(None))) \
                .order_by(desc(SampleEna.collection_date)).first()
        elif source == DataSource.GISAID:
            result = self.session.query(SampleGisaid.date).filter(
                and_(SampleGisaid.finished, SampleGisaid.date.isnot(None))) \
                .order_by(desc(SampleGisaid.date)).first()
        else:
            raise CovigatorQueryException("No valid data source for query of most recent sample")
        return result[0] if result is not None else result

    def get_date_of_last_check(self, data_source: DataSource) -> date:
        """
        Returns the date of the latest non failed accessor check that also has a subsequent non failed processor run.
        Until the processor has ran the data fetched from the accessor is not available.
        """
        result1 = self.session.query(Log.start).filter(
            and_(Log.source == data_source, Log.module == CovigatorModule.PROCESSOR, Log.has_error == False)).order_by(
            desc(Log.start)).first()
        most_recent_processor_run = result1[0] if result1 is not None else result1
        result2 = None
        if most_recent_processor_run:
            result2 = self.session.query(Log.start).filter(
                and_(Log.source == data_source, Log.module == CovigatorModule.ACCESSOR, Log.has_error == False,
                     Log.start < most_recent_processor_run)).order_by(desc(Log.start)).first()
        return result2[0] if result2 is not None else result2

    def get_date_of_last_update(self, data_source: DataSource) -> date:
        """
        Returns the date of the latest non failed accessor check **with some new data** that also has a subsequent non
        failed processor run.
        Until the processor has ran the data fetched from the accessor is not available.
        """
        result1 = self.session.query(Log.start).filter(
            and_(Log.source == data_source, Log.module == CovigatorModule.PROCESSOR, Log.has_error == False)).order_by(
            desc(Log.start)).first()
        most_recent_processor_run = result1[0] if result1 is not None else result1
        result2 = None
        if most_recent_processor_run:
            result2 = self.session.query(Log.start).filter(
                and_(Log.source == data_source, Log.module == CovigatorModule.ACCESSOR, Log.has_error == False,
                     Log.processed > 0, Log.start < most_recent_processor_run)) \
                .order_by(desc(Log.start)).first()
        return result2[0] if result2 is not None else result2

    def get_top_occurring_variants(self, top, source: str = None):
        query = self.session.query(
            VariantObservation.variant_id, VariantObservation.hgvs_p, VariantObservation.gene_name,
            VariantObservation.annotation_highest_impact, func.count().label('total')) \
            .filter(VariantObservation.annotation_highest_impact != SYNONYMOUS_VARIANT)
        if source is not None:
            query = query.filter(VariantObservation.source == source)
        query = query.group_by(VariantObservation.variant_id, VariantObservation.hgvs_p, VariantObservation.gene_name,
                      VariantObservation.annotation_highest_impact) \
            .order_by(desc('total')).limit(top)
        top_occurring_variants = pd.read_sql(query.statement, self.session.bind)

        # calculate frequency
        count_samples = self.count_samples(source=source if source is not None else None)
        top_occurring_variants['frequency'] = top_occurring_variants.total.transform(
            lambda x: round(float(x) / count_samples, 3))

        # add counts for every month
        top_occurring_variants = self._get_counts_per_month(top_occurring_variants=top_occurring_variants, source=source)

        return top_occurring_variants

    def _get_counts_per_month(self, top_occurring_variants, source=None):
        variant_counts_by_month = []
        for _, variant in top_occurring_variants.iterrows():
            variant_counts_by_month.append(self.get_variant_counts_by_month(variant.variant_id, source=source))
        top_occurring_variants_by_month = pd.concat(variant_counts_by_month)
        # get total count of samples per month to calculate the frequency by month
        sample_counts_by_month = self.get_sample_counts_by_month(source=source)
        top_occurring_variants_by_month = pd.merge(
            left=top_occurring_variants_by_month, right=sample_counts_by_month, how="left", on="month")
        top_occurring_variants_by_month["frequency_by_month"] = \
            (top_occurring_variants_by_month["count"] / top_occurring_variants_by_month["sample_count"]). \
                transform(lambda x: round(x, 3))
        # join both tables with total counts and counts per month
        top_occurring_variants = pd.merge(
            left=top_occurring_variants, right=top_occurring_variants_by_month, on="variant_id", how="left")
        # format the month column appropriately
        top_occurring_variants.month = top_occurring_variants.month.transform(
            lambda d: "{}-{:02d}".format(d.year, int(d.month)))
        return top_occurring_variants

    def get_top_occurring_variants_precomputed(self, top=10, gene_name=None, metric="count", source=None) -> pd.DataFrame:
        """
        Returns the top occurring variants + the segregated counts of occurrences per month
        with columns: chromosome, position, reference, alternate, total, month, count
        """
        query = self.session.query(PrecomputedOccurrence).filter(PrecomputedOccurrence.source == source)
        if gene_name is not None:
            query = query.filter(PrecomputedOccurrence.gene_name == gene_name)
        if metric == "count":
            query = query.order_by(PrecomputedOccurrence.count.desc())
        elif metric == "frequency":
            query = query.order_by(PrecomputedOccurrence.frequency.desc())
        else:
            raise CovigatorQueryException("Not supported metric for top occurring variants")

        top_occurring_variants = pd.read_sql(query.statement, self.session.bind)

        # formats the DNA mutation
        top_occurring_variants.rename(columns={'variant_id': 'dna_mutation'}, inplace=True)

        # pivots the table over months
        top_occurring_variants = pd.pivot_table(
            top_occurring_variants, index=['gene_name', 'dna_mutation', 'hgvs_p', 'annotation', "frequency", "total"],
            columns=["month"], values=[metric], fill_value=0).droplevel(0, axis=1).reset_index()

        return top_occurring_variants.sort_values(by="frequency", ascending=False).head(top)

    def get_variant_counts_by_month(self, variant_id, source=None) -> pd.DataFrame:

        sql_query_ds_ena = """
        select count(*) as count, variant_id, date_trunc('month', date::timestamp) as month 
            from {variant_observation_table} 
            where variant_id='{variant_id}' {source_filter}
            group by variant_id, date_trunc('month', date::timestamp);
            """.format(
            variant_observation_table=VARIANT_OBSERVATION_TABLE_NAME,
            variant_id=variant_id,
            source_filter="and source='{source}'".format(source=source) if source is not None else ""
        )
        data = pd.read_sql_query(sql_query_ds_ena, self.session.bind)
        return data[data.month is not None]

    def get_sample_counts_by_month(self, source=None) -> pd.DataFrame:
        counts_ena = None
        if source is None or source == DataSource.ENA:
            query = self.session.query(
                func.date_trunc('month', SampleEna.collection_date).label("month"),
                func.count().label("sample_count"))\
                .filter(SampleEna.finished) \
                .group_by(func.date_trunc('month', SampleEna.collection_date))
            counts_ena = pd.read_sql(query.statement, self.session.bind)
        counts_gisaid = None
        if source is None or source == DataSource.GISAID:
            query = self.session.query(
                func.date_trunc('month', SampleGisaid.date).label("month"),
                func.count().label("sample_count"))\
                .filter(SampleGisaid.finished) \
                .group_by(func.date_trunc('month', SampleGisaid.date))
            counts_gisaid = pd.read_sql(query.statement, self.session.bind)
        if counts_gisaid is None:
            # NOTE: setting index and then resetting is necessary to get the date column in the right dtype
            counts = counts_ena
        elif counts_ena is None:
            # NOTE: setting index and then resetting is necessary to get the date column in the right dtype
            counts = counts_gisaid
        else:
            counts = counts_gisaid.set_index(["month"]).add(
                counts_ena.set_index(["month"]), fill_value=0).reset_index()
        return counts

    def get_variants_cooccurrence_by_gene(self, gene_name, min_cooccurrence=5, test=False) -> pd.DataFrame:
        """
        Returns the full cooccurrence matrix of all non synonymous variants in a gene with at least
        min_occurrences occurrences.
        """
        # query for total samples required to calculate frequencies
        # FIXME: once we have the cooccurrence for GISAID samples we need to count all samples here
        count_samples = self.count_samples(source=DataSource.ENA.name)

        # query for cooccurrence matrix
        variant_one = aliased(Variant)
        variant_two = aliased(Variant)
        if not test:
            query = self.session.query(VariantCooccurrence,
                                       variant_one.position,
                                       variant_one.reference,
                                       variant_one.alternate,
                                       variant_one.hgvs_p,
                                       (variant_one.hgvs_p + " - " + variant_two.hgvs_p).label("hgvs_tooltip"))
        else:
            # this is needed for testing environment with SQLite
            query = self.session.query(VariantCooccurrence,
                                       variant_one.position,
                                       variant_one.reference,
                                       variant_one.alternate,
                                       variant_one.hgvs_p,
                                       (variant_one.hgvs_p + " - " + variant_two.hgvs_p).label("hgvs_tooltip"))

        query = query.filter(VariantCooccurrence.count >= min_cooccurrence)\
            .join(variant_one, and_(VariantCooccurrence.variant_id_one == variant_one.variant_id)) \
            .join(variant_two, and_(VariantCooccurrence.variant_id_two == variant_two.variant_id))

        if gene_name is not None:
            query = query.filter(and_(variant_one.gene_name == gene_name,
                         variant_one.annotation != SYNONYMOUS_VARIANT,
                         variant_two.gene_name == gene_name,
                         variant_two.annotation != SYNONYMOUS_VARIANT))
        else:
            query = query.filter(and_(variant_one.annotation != SYNONYMOUS_VARIANT,
                                      variant_two.annotation != SYNONYMOUS_VARIANT))

        self._print_query(query=query)
        data = pd.read_sql(query.statement, self.session.bind)

        full_matrix = None
        if data.shape[0] > 0:
            # these are views of the original data
            annotations = data.loc[data.variant_id_one == data.variant_id_two,
                                   ["variant_id_one", "position", "reference", "alternate", "hgvs_p"]]
            tooltip = data.loc[:, ["variant_id_one", "variant_id_two", "hgvs_tooltip"]]
            diagonal = data.loc[data.variant_id_one == data.variant_id_two, ["variant_id_one", "variant_id_two", "count"]]
            sparse_matrix = data.loc[:, ["variant_id_one", "variant_id_two", "count"]]
            sparse_matrix["frequency"] = sparse_matrix["count"] / count_samples
            sparse_matrix = pd.merge(left=sparse_matrix, right=diagonal, on="variant_id_one", how="left", suffixes=("", "_one"))
            sparse_matrix = pd.merge(left=sparse_matrix, right=diagonal, on="variant_id_two", how="left", suffixes=("", "_two"))

            # calculate Jaccard index
            sparse_matrix["count_union"] = sparse_matrix["count_one"] + sparse_matrix["count_two"] - sparse_matrix["count"]
            sparse_matrix["jaccard"] = sparse_matrix["count"] / sparse_matrix["count_union"]

            # calculate Cohen's kappa
            sparse_matrix["chance_agreement"] = np.exp(-sparse_matrix["count"])
            sparse_matrix["kappa"] = 1 - ((1 - sparse_matrix.jaccard) / (1 - sparse_matrix.chance_agreement))
            sparse_matrix["kappa"] = sparse_matrix["kappa"].transform(lambda k: k if k > 0 else 0)

            del sparse_matrix["count_union"]
            del sparse_matrix["count_one"]
            del sparse_matrix["count_two"]
            del sparse_matrix["chance_agreement"]

            # from the sparse matrix builds in memory the complete matrix
            all_variants = data.variant_id_one.unique()
            empty_full_matrix = pd.DataFrame(index=pd.MultiIndex.from_product(
                [all_variants, all_variants], names=["variant_id_one", "variant_id_two"])).reset_index()
            full_matrix = pd.merge(
                left=empty_full_matrix, right=sparse_matrix, on=["variant_id_one", "variant_id_two"], how='left')

            # add annotation on variant one
            full_matrix = pd.merge(left=full_matrix, right=annotations, on="variant_id_one", how='left')\
                .rename(columns={"position": "position_one",
                                 "reference": "reference_one",
                                 "alternate": "alternate_one",
                                 "hgvs_p": "hgvs_p_one"})

            # add annotations on variant two
            full_matrix = pd.merge(left=full_matrix, right=annotations,
                                   left_on="variant_id_two", right_on="variant_id_one", how='left') \
                .rename(columns={"variant_id_one_x": "variant_id_one",
                                 "position": "position_two",
                                 "reference": "reference_two",
                                 "alternate": "alternate_two",
                                 "hgvs_p": "hgvs_p_two"})

            # add tooltip
            full_matrix = pd.merge(left=full_matrix, right=tooltip, on=["variant_id_one", "variant_id_two"], how='left')
            # correct tooltip in diagonal
            full_matrix.loc[full_matrix.variant_id_one == full_matrix.variant_id_two, "hgvs_tooltip"] = full_matrix.hgvs_p_one

            # NOTE: transpose matrix manually as plotly transpose does not work with labels
            # the database return the upper diagonal, the lower is best for plots
            full_matrix.sort_values(["position_two", "reference_two", "alternate_two",
                                     "position_one", "reference_one", "alternate_one"], inplace=True)

            full_matrix = full_matrix.loc[:, ["variant_id_one", "variant_id_two", "count", "frequency", "jaccard",
                                              "kappa", "hgvs_tooltip"]]

        return full_matrix

    @functools.lru_cache()
    def get_mds(self, gene_name, min_cooccurrence, min_samples) -> pd.DataFrame:

        variant_one = aliased(Variant)
        variant_two = aliased(Variant)
        query = self.session.query(VariantCooccurrence,
                                   variant_one.hgvs_p.label("hgvs_p_one"),
                                   variant_two.hgvs_p.label("hgvs_p_two")) \
            .filter(VariantCooccurrence.count >= min_cooccurrence) \
            .join(variant_one, and_(VariantCooccurrence.variant_id_one == variant_one.variant_id)) \
            .join(variant_two, and_(VariantCooccurrence.variant_id_two == variant_two.variant_id))

        if gene_name is not None:
            query = query.filter(and_(variant_one.gene_name == gene_name,
                                      variant_one.annotation != SYNONYMOUS_VARIANT,
                                      variant_two.gene_name == gene_name,
                                      variant_two.annotation != SYNONYMOUS_VARIANT))
        else:
            query = query.filter(and_(variant_one.annotation != SYNONYMOUS_VARIANT,
                                      variant_two.annotation != SYNONYMOUS_VARIANT))

        sparse_matrix = pd.read_sql(query.statement, self.session.bind)
        diagonal = sparse_matrix.loc[sparse_matrix.variant_id_one == sparse_matrix.variant_id_two,
                                     ["variant_id_one", "variant_id_two", "count"]]
        sparse_matrix_with_diagonal = pd.merge(
            left=sparse_matrix, right=diagonal, on="variant_id_one", how="left", suffixes=("", "_one"))
        sparse_matrix_with_diagonal = pd.merge(
            left=sparse_matrix_with_diagonal, right=diagonal, on="variant_id_two", how="left", suffixes=("", "_two"))

        # calculate Jaccard index
        sparse_matrix_with_diagonal["count_union"] = sparse_matrix_with_diagonal["count_one"] + \
                                                     sparse_matrix_with_diagonal["count_two"] - \
                                                     sparse_matrix_with_diagonal["count"]
        sparse_matrix_with_diagonal["jaccard_similarity"] = sparse_matrix_with_diagonal["count"] / \
                                                               sparse_matrix_with_diagonal.count_union
        sparse_matrix_with_diagonal["jaccard_dissimilarity"] = 1 - sparse_matrix_with_diagonal.jaccard_similarity

        # calculate Cohen's kappa
        sparse_matrix_with_diagonal["chance_agreement"] = np.exp(-sparse_matrix_with_diagonal["count"])
        sparse_matrix_with_diagonal["kappa"] = 1 - ((1 - sparse_matrix_with_diagonal.jaccard_similarity) / (
                1 - sparse_matrix_with_diagonal.chance_agreement))
        sparse_matrix_with_diagonal["kappa"] = sparse_matrix_with_diagonal["kappa"].transform(lambda k: k if k > 0 else 0)
        sparse_matrix_with_diagonal["kappa_dissimilarity"] = 1 - sparse_matrix_with_diagonal.kappa

        dissimilarity_metric = "kappa_dissimilarity"

        # build upper diagonal matrix
        all_variants = sparse_matrix_with_diagonal.variant_id_one.unique()
        empty_full_matrix = pd.DataFrame(index=pd.MultiIndex.from_product(
            [all_variants, all_variants], names=["variant_id_one", "variant_id_two"])).reset_index()
        upper_diagonal_matrix = pd.merge(
            # gets only the inferior matrix without the diagnonal
            left=empty_full_matrix.loc[empty_full_matrix.variant_id_one < empty_full_matrix.variant_id_two, :],
            right=sparse_matrix_with_diagonal.loc[:, ["variant_id_one", "variant_id_two", dissimilarity_metric]],
            on=["variant_id_one", "variant_id_two"], how='left')
        upper_diagonal_matrix.fillna(1.0, inplace=True)
        upper_diagonal_matrix.sort_values(by=["variant_id_one", "variant_id_two"], inplace=True)

        logger.info("Building square distance matrix...")
        distance_matrix = squareform(upper_diagonal_matrix[dissimilarity_metric])
        # this ensures the order of variants ids is coherent with the non redundant form of the distance matrix
        ids = np.array([list(upper_diagonal_matrix.variant_id_one[0])[0]] + \
              list(upper_diagonal_matrix.variant_id_two[0:len(upper_diagonal_matrix.variant_id_two.unique())]))
        distance_matrix_with_ids = DissimilarityMatrix(data=distance_matrix, ids=ids)

        logger.info("Clustering...")
        clusters = OPTICS(min_samples=min_samples, max_eps=1.4).fit_predict(distance_matrix_with_ids.data)

        logger.info("Building clustering dataframe...")
        data = pd.DataFrame({'variant_id': distance_matrix_with_ids.ids, 'cluster': clusters})

        logger.info("Annotate with HGVS.p ...")
        annotations = pd.concat([
            sparse_matrix.loc[:, ["variant_id_one", "hgvs_p_one"]].rename(
                columns={"variant_id_one": "variant_id", "hgvs_p_one": "hgvs_p"}),
            sparse_matrix.loc[:, ["variant_id_two", "hgvs_p_two"]].rename(
                columns={"variant_id_two": "variant_id", "hgvs_p_two": "hgvs_p"})])
        data = pd.merge(left=data, right=annotations, on="variant_id", how="left")

        logger.info("Annotate with cluster mean Jaccard index...")
        data["cluster_jaccard_mean"] = 1.0
        for c in data.cluster.unique():
            variants_in_cluster = data[data.cluster == c].variant_id.unique()
            data.cluster_jaccard_mean = np.where(data.cluster == c, sparse_matrix_with_diagonal[
                (sparse_matrix.variant_id_one.isin(variants_in_cluster)) &
                (sparse_matrix.variant_id_two.isin(variants_in_cluster)) &
                (sparse_matrix.variant_id_one != sparse_matrix.variant_id_two)
            ].jaccard_dissimilarity.mean(), data.cluster_jaccard_mean)

        data.cluster_jaccard_mean = data.cluster_jaccard_mean.transform(lambda x: round(x, 3))

        data = data.set_index("variant_id").drop_duplicates().reset_index().sort_values("cluster")

        return data[data.cluster != -1].loc[:, ["cluster", "variant_id", "hgvs_p", "cluster_jaccard_mean"]]

    def get_variant_abundance_histogram(self, bin_size=50, source: str = None, cache=True) -> pd.DataFrame:
        histogram = None
        if cache:
            query = self.session.query(PrecomputedVariantAbundanceHistogram)
            if source is not None:
                query = query.filter(and_(PrecomputedVariantAbundanceHistogram.bin_size == bin_size,
                                          PrecomputedVariantAbundanceHistogram.source == source))
            else:
                query = query.filter(and_(PrecomputedVariantAbundanceHistogram.bin_size == bin_size,
                                          PrecomputedVariantAbundanceHistogram.source == None))
            histogram = pd.read_sql(query.statement, self.session.bind)
            if histogram.shape[0] == 0:
                raise CovigatorDashboardMissingPrecomputedData
            histogram = histogram[["position_bin", "count_unique_variants", "count_variant_observations"]]
        else:
            # queries for the maximum position
            maximum_position = self.session.query(func.max(Variant.position)).first()[0]
            if maximum_position is not None:
                # builds all possible bins
                all_bins = pd.DataFrame(data=[i*bin_size for i in range(int(maximum_position/bin_size) + 1)], columns=["position_bin"])

                # counts variants over those bins
                sql_query = """
                        SELECT cast("position"/{bin_size} as int)*{bin_size} AS position_bin,
                               COUNT(*) as count_unique_variants
                        FROM {table_name}
                        GROUP BY position_bin
                        ORDER BY position_bin;
                        """.format(bin_size=bin_size, table_name=Variant.__tablename__)
                binned_counts_variants = pd.read_sql_query(sql_query, self.session.bind)

                # counts variant observations over those bins
                sql_query = """
                        SELECT cast("position"/{bin_size} as int)*{bin_size} AS position_bin,
                               COUNT(*) as count_variant_observations
                        FROM {table_name}
                        {source_filter}
                        GROUP BY position_bin
                        ORDER BY position_bin;
                        """.format(bin_size=bin_size, table_name=VariantObservation.__tablename__,
                                   source_filter="WHERE source='{}'".format(source) if source is not None else "")
                binned_counts_variant_observations = pd.read_sql_query(sql_query, self.session.bind)

                histogram = pd.merge(
                    left=pd.merge(
                        left=all_bins,
                        right=binned_counts_variants,
                        on="position_bin", how="left").reset_index(),
                    right=binned_counts_variant_observations,
                    on="position_bin", how="left").reset_index()
                histogram.fillna(0, inplace=True)
                histogram = histogram[["position_bin", "count_unique_variants", "count_variant_observations"]]

        return histogram

    def get_conservation_table(self, bin_size=50, start=None, end=None) -> pd.DataFrame:
        # counts variants over those bins
        sql_query = """
                SELECT cast("start"/{bin_size} as int)*{bin_size} AS position_bin,
                       AVG("conservation") as conservation,
                       AVG("conservation_sarbecovirus") as conservation_sarbecovirus,
                       AVG("conservation_vertebrates") as conservation_vertebrates
                FROM {table_name}
                {where}
                GROUP BY position_bin
                ORDER BY position_bin;
                """.format(bin_size=bin_size, table_name= Conservation.__tablename__,
                           where="WHERE start >= {start} and start <= {end}".format(start=start, end=end)
                           if start is not None and end is not None else "")
        return pd.read_sql_query(sql_query, self.session.bind)

    def _print_query(self, query):
        class StringLiteral(String):
            """Teach SA how to literalize various things."""

            def literal_processor(self, dialect):
                super_processor = super(StringLiteral, self).literal_processor(dialect)

                def process(value):
                    if isinstance(value, int):
                        return str(value)
                    if not isinstance(value, str):
                        value = str(value)
                    result = super_processor(value)
                    if isinstance(result, bytes):
                        result = result.decode(dialect.encoding)
                    return result

                return process

        class LiteralDialect(DefaultDialect):
            colspecs = {
                # prevent various encoding explosions
                String: StringLiteral,
                # teach SA about how to literalize a datetime
                DateTime: StringLiteral,
                # don't format py2 long integers to NULL
                NullType: StringLiteral,
            }
        logger.info(query.statement.compile(
            dialect=LiteralDialect(),
            compile_kwargs={'literal_binds': True}).string)



from typing import Tuple
from faker import Faker
from sqlalchemy.orm import Session

from covigator.database.model import SampleEna, Sample, DataSource, JobEna, JobStatus, Log, CovigatorModule, Variant, \
    VariantObservation, VariantCooccurrence, VariantType
from Bio.Alphabet.IUPAC import IUPACData


def get_mocked_variant(faker: Faker, chromosome=None, gene_name=None) -> Variant:
    variant = Variant(
        chromosome=chromosome if chromosome else faker.bothify(text="chr##"),
        position=faker.random_int(min=1, max=30000),
        reference=faker.random_choices(list(IUPACData.unambiguous_dna_letters), length=1)[0],
        # TODO: reference and alternate could be equal!
        alternate=faker.random_choices(list(IUPACData.unambiguous_dna_letters), length=1)[0],
        variant_type=VariantType.SNV,
        gene_name=gene_name,
        hgvs_p="p.{}{}{}".format(
            faker.random_choices(list(IUPACData.protein_letters_1to3.values()), length=1)[0],
            faker.random_int(min=1, max=500),
            faker.random_choices(list(IUPACData.protein_letters_1to3.values()), length=1)[0]
        ),
        annotation=faker.random_choices(["non_synonymous", "inframe_deletion", "inframe_insertion"], length=1)[0]
    )
    variant.variant_id = variant.get_variant_id()
    return variant


def get_mocked_variant_observation(sample: Sample, variant: Variant, faker=Faker()):
    return VariantObservation(
        sample=sample.id if sample else faker.unique.uuid4(),
        source=sample.source if sample else faker.random_choices((DataSource.ENA, DataSource.GISAID)),
        variant_id=variant.variant_id,
        chromosome=variant.chromosome,
        position=variant.position,
        reference=variant.reference,
        alternate=variant.alternate,
        variant_type=VariantType.SNV)


def get_mocked_ena_sample(faker: Faker, job_status=JobStatus.FINISHED) -> Tuple[SampleEna, Sample, JobEna]:
    """
    Returns a triple of SampleEna, Sample and Job with the same sample identifier
    """
    identifier = faker.unique.uuid4()
    sample_ena = SampleEna(
        run_accession=identifier,
        collection_date=faker.date_time(),
        country=faker.country(),
        fastq_ftp=faker.uri(),
        fastq_md5=faker.md5(),
        num_fastqs=1,
        finished=job_status == JobStatus.FINISHED
    )
    sample = Sample(
        id=identifier,
        source=DataSource.ENA,
        ena_id=identifier
    )
    job = JobEna(
        run_accession=identifier,
        status=job_status
    )
    return sample_ena, sample, job


def get_mocked_log(faker: Faker, source: DataSource = None, module: CovigatorModule = None) -> Log:
    return Log(
        start=faker.date_time(),
        end=faker.date_time(),
        source=source if source else faker.random_choices((DataSource.ENA, DataSource.GISAID), length=1)[0],
        module=module if module else faker.random_choices((CovigatorModule.ACCESSOR, CovigatorModule.PROCESSOR),
                                                          length=1)[0],
        has_error=faker.boolean(),
        processed=faker.random_digit(),
        data={"included": faker.random_digit(),
              "excluded": {"this": faker.random_digit(), "that": faker.random_digit()}}
    )


def get_mocked_variant_cooccurrence(faker: Faker, variant_one: Variant, variant_two: Variant) -> VariantCooccurrence:
    if variant_one.position <= variant_two.position:
        cooccurrence = VariantCooccurrence(
            variant_id_one=variant_one.variant_id,
            variant_id_two=variant_two.variant_id,
            count=faker.random_int(min=1, max=10)
        )
    else:
        cooccurrence = VariantCooccurrence(
            variant_id_two=variant_one.variant_id,
            variant_id_one=variant_two.variant_id,
            count=faker.random_int(min=1, max=10)
        )
    return cooccurrence


def mock_samples_and_variants(faker, session: Session, num_samples=10):
    existing_variants = set()
    for _ in range(num_samples):
        sample_ena, sample, job = get_mocked_ena_sample(faker=faker)
        session.add_all([sample_ena, sample, job])
        variants = [get_mocked_variant(faker=faker) for _ in range(10)]
        session.add_all(list(filter(lambda x: x in existing_variants, variants)))
        existing_variants.update([v.variant_id for v in variants])
        session.commit()
        variants_observations = [get_mocked_variant_observation(faker=faker, variant=v, sample=sample)
                                 for v in variants]
        session.add_all(variants_observations)
    session.commit()



from datetime import date, datetime
import requests
from sqlalchemy.orm import Session

from covigator.accessor import MINIMUM_DATE
from covigator.exceptions import CovigatorExcludedSampleTooEarlyDateException
from covigator.misc import backoff_retrier
from covigator.database.model import SampleEna, JobEna, Sample, DataSource, Log, CovigatorModule
from covigator.database.database import Database
from logzero import logger
from covigator.misc.country_parser import CountryParser

NUMBER_RETRIES = 5


class EnaAccessor:

    ENA_API_URL_BASE = "https://www.ebi.ac.uk/ena/portal/api"
    PAGE_SIZE = 1000
    # see https://www.ebi.ac.uk/ena/portal/api/returnFields?result=read_run&format=json for all possible fields
    ENA_FIELDS = [
        # data on run
        "scientific_name",
        "study_accession",
        "experiment_accession",
        "first_created",
        "collection_date",
        "instrument_platform",
        "instrument_model",
        "library_name",
        "nominal_length",
        "library_layout",
        "library_strategy",
        "library_source",
        "library_selection",
        "read_count",
        "base_count",
        "sample_collection",
        "sequencing_method",
        "center_name",
        # FASTQs
        "fastq_ftp",
        "fastq_md5",
        # data on host
        "host_tax_id",
        "host_sex",
        "host_body_site",
        "host_gravidity",
        "host_phenotype",
        "host_genotype",
        # geographical data
        "lat",
        "lon",
        "country"
    ]
    INCLUDED_LIBRARY_STRATEGIES = [
        "WGA",
        "WGS",
        "Targeted-Capture",
        "RNA-Seq"
    ]

    def __init__(self, tax_id: str, host_tax_id: str, database: Database, maximum=None):
        logger.info("Initialising ENA accessor")
        self.start_time = datetime.now()
        self.has_error = False
        self.error_message = None
        self.tax_id = tax_id
        assert self.tax_id is not None and self.tax_id.strip() != "", "Empty tax id"
        logger.info("Tax id {}".format(self.tax_id))
        self.host_tax_id = host_tax_id
        assert self.host_tax_id is not None and self.host_tax_id.strip() != "", "Empty host tax id"
        logger.info("Host tax id {}".format(self.host_tax_id))
        self.database = database
        assert self.database is not None, "Empty database"
        self.maximum = maximum

        self.excluded_samples_by_host_tax_id = {}
        self.excluded_samples_by_fastq_ftp = 0
        self.excluded_samples_by_instrument_platform = {}
        self.excluded_samples_by_library_strategy = {}
        self.excluded_existing = 0
        self.included = 0
        self.excluded = 0
        self.excluded_by_date = 0
        self.country_parser = CountryParser()

        # this ensures there is a retry mechanism in place with a limited number of retries
        self.get_with_retries = backoff_retrier.wrapper(requests.get, NUMBER_RETRIES)

    def access(self):
        logger.info("Starting ENA accessor")
        offset = 0
        finished = False
        session = self.database.get_database_session()
        # NOTE: holding in memory the whole list of existing ids is much faster than querying every time
        # it assumes there will be no repetitions
        existing_sample_ids = [value for value, in session.query(SampleEna.run_accession).all()]
        try:
            while not finished:
                list_runs = self._get_ena_runs_page(offset)
                logger.info("Read page of {} ENA samples".format(len(list_runs)))
                self._process_runs(list_runs, existing_sample_ids, session)
                # finishes when no more data or when test parameter maximum is reached
                if len(list_runs) < self.PAGE_SIZE or (self.maximum is not None and self.included >= self.maximum):
                    finished = True
                offset += len(list_runs)
        except Exception as e:
            logger.exception(e)
            session.rollback()
            self.has_error = True
            self.error_message = str(e)
        finally:
            self._write_execution_log(session)
            session.close()
            self._log_results()
            logger.info("Finished ENA accessor")

    def _get_ena_runs_page(self, offset):
        return self.get_with_retries(
            "{url_base}/search?result=read_run&"
            "query=tax_eq({tax_id})&"
            "limit={page_size}&"
            "offset={offset}&"
            "fields={fields}&"
            "format=json".format(
                url_base=self.ENA_API_URL_BASE,
                tax_id=self.tax_id,
                page_size=self.PAGE_SIZE,
                offset=offset,
                fields=",".join(self.ENA_FIELDS)
            )).json()

    def _process_runs(self, list_runs, existing_sample_ids, session: Session):

        included_samples = []
        included_samples_ena = []
        included_jobs = []
        for run in list_runs:
            if isinstance(run, dict):
                if run.get("run_accession") in existing_sample_ids:
                    self.excluded_existing += 1
                    continue    # skips runs already registered in the database
                if not self._complies_with_inclusion_criteria(run):
                    continue    # skips runs not complying with inclusion criteria
                # NOTE: this parse operation is costly
                try:
                    sample_ena = self._parse_ena_run(run)
                    sample = self._build_sample(sample_ena)
                    self.included += 1
                    included_samples_ena.append(sample_ena)
                    included_samples.append(sample)
                    included_jobs.append(JobEna(run_accession=sample_ena.run_accession))
                except CovigatorExcludedSampleTooEarlyDateException:
                    logger.error("Excluded sample due to too early date")
                    self.excluded_by_date += 1
                    self.excluded += 1
            else:
                logger.error("Run from ENA without the expected format")

        if len(included_samples) > 0:
            session.add_all(included_samples_ena)
            session.commit()
            session.add_all(included_samples)
            session.add_all(included_jobs)
            session.commit()
            logger.info("Added {} new ENA samples".format(len(included_samples)))
        logger.info("Processed {} ENA samples".format(len(list_runs)))

    def _build_sample(self, sample_ena):
        return Sample(
            id=sample_ena.run_accession,
            source=DataSource.ENA,
            ena_id=sample_ena.run_accession
        )

    def _parse_country(self, sample: SampleEna):
        parsed_country = self.country_parser.parse_country(
            sample.country.split(":")[0] if sample.country else "")
        sample.country_raw = sample.country
        sample.country = parsed_country.country
        sample.country_alpha_2 = parsed_country.country_alpha_2
        sample.country_alpha_3 = parsed_country.country_alpha_3
        sample.continent_alpha_2 = parsed_country.continent_alpha_2
        sample.continent = parsed_country.continent

    def _parse_dates(self, ena_run):
        ena_run.collection_date = self._parse_abstract(ena_run.collection_date, date.fromisoformat)
        ena_run.first_created = self._parse_abstract(ena_run.first_created, date.fromisoformat)
        if ena_run.first_created is not None and ena_run.first_created < MINIMUM_DATE:
            raise CovigatorExcludedSampleTooEarlyDateException

    def _parse_ena_run(self, run):
        sample = SampleEna(**run)
        self._parse_country(sample)
        self._parse_dates(sample)
        self._parse_numeric_fields(sample)
        fastqs = sample.get_fastqs_ftp()
        # annotates with the number of FASTQ files, this is useful as we hold the FASTQs in a single string
        sample.num_fastqs = 0 if fastqs is None or fastqs == [""] else len(fastqs)
        return sample

    def _parse_numeric_fields(self, ena_run):
        ena_run.nominal_length = self._parse_abstract(ena_run.nominal_length, int)
        ena_run.read_count = self._parse_abstract(ena_run.read_count, int)
        ena_run.base_count = self._parse_abstract(ena_run.base_count, int)
        ena_run.lat = self._parse_abstract(ena_run.lat, float)
        ena_run.lon = self._parse_abstract(ena_run.lon, float)

    def _parse_abstract(self, value, type):
        try:
            value = type(value)
        except (ValueError, TypeError):
            value = None
        return value

    def _complies_with_inclusion_criteria(self, ena_run: dict):
        # NOTE: this uses the original dictionary instead of the parsed SampleEna class for performance reasons
        included = True
        host_tax_id = ena_run.get("host_tax_id")
        if host_tax_id is None or host_tax_id.strip() == "" or host_tax_id != self.host_tax_id:
            included = False    # skips runs where the host is empty or does not match
            self.excluded_samples_by_host_tax_id[str(host_tax_id)] = \
                self.excluded_samples_by_host_tax_id.get(str(host_tax_id), 0) + 1
        fastq_ftp = ena_run.get("fastq_ftp")
        if fastq_ftp is None or fastq_ftp == "":
            included = False    # skips runs without FTP URL
            self.excluded_samples_by_fastq_ftp += 1
        instrument_platform = ena_run.get("instrument_platform")
        if instrument_platform.upper() != "ILLUMINA":
            included = False    # skips non Illumina data
            self.excluded_samples_by_instrument_platform[str(instrument_platform)] = \
                self.excluded_samples_by_instrument_platform.get(str(instrument_platform), 0) + 1
        library_strategy = ena_run.get("library_strategy")
        if library_strategy not in self.INCLUDED_LIBRARY_STRATEGIES:
            included = False  # skips not included library strategies
            self.excluded_samples_by_library_strategy[str(library_strategy)] = \
                self.excluded_samples_by_library_strategy.get(str(library_strategy), 0) + 1
        if not included:
            self.excluded += 1
        return included

    def _log_results(self):
        logger.info("Included new runs = {}".format(self.included))
        logger.info("Excluded already existing runs = {}".format(self.excluded_existing))
        logger.info("Total excluded runs by selection criteria = {}".format(self.excluded))
        logger.info("Excluded due to empty FASTQ FTP URL runs = {}".format(self.excluded_samples_by_fastq_ftp))
        logger.info("Excluded by platform runs = {}".format(self.excluded_samples_by_instrument_platform))
        logger.info("Excluded by host if runs = {}".format(self.excluded_samples_by_host_tax_id))
        logger.info("Excluded by library strategy = {}".format(self.excluded_samples_by_library_strategy))

    def _write_execution_log(self, session: Session):
        end_time = datetime.now()
        session.add(Log(
            start=self.start_time,
            end=end_time,
            source=DataSource.ENA,
            module=CovigatorModule.ACCESSOR,
            has_error=self.has_error,
            error_message=self.error_message,
            processed=self.included,
            data={
                "included": self.included,
                "excluded": {
                    "existing": self.excluded_existing,
                    "excluded_by_criteria": self.excluded,
                    "excluded_by_date": self.excluded_by_date,
                    "missing_fastq": self.excluded_samples_by_fastq_ftp,
                    "platform": self.excluded_samples_by_instrument_platform,
                    "library_strategy": self.excluded_samples_by_library_strategy,
                    "host": self.excluded_samples_by_host_tax_id
                }
            }
        ))
        session.commit()



from unittest import TestCase
from covigator.database.database import Database
from covigator.database.model import Gene, get_table_versioned_name, Variant, Conservation
import pandas as pd
from covigator.configuration import Configuration
from covigator.tests.unit_tests.faked_objects import FakeConfiguration


class DatabaseInitialisationTests(TestCase):

    def setUp(self) -> None:
        self.config = FakeConfiguration()

    def test_genes_table_initialisation(self):
        database = Database(test=True, config=self.config)
        session = database.get_database_session()
        self.assertGreater(session.query(Gene).count(), 0)
        for g in session.query(Gene).all():
            self.assertIsNotNone(g.identifier)
            self.assertIsNotNone(g.name)
            self.assertTrue(g.data is not None or g.sequence is not None)

    def test_genes_table_initialisation_not_twice(self):
        database = Database(test=True, config=self.config)
        session = database.get_database_session()
        count_genes = session.query(Gene).count()

        # creates another connection
        database2 = Database(test=True, config=self.config)
        session2 = database2.get_database_session()

        count_genes_2 = session2.query(Gene).count()
        self.assertEqual(count_genes, count_genes_2)

    def test_versioned_tables(self):
        config = Configuration()
        config.db_table_version = "_v1"
        self.assertEqual("gene_v1", get_table_versioned_name(Gene.__table__.name, config=config))
        config.db_table_version = "_v2"
        self.assertEqual("variant_v2", get_table_versioned_name(Variant.__table__.name, config=config))

    def test_conservation_loader(self):
        database = Database(test=True, config=self.config)
        session = database.get_database_session()
        self.assertGreater(session.query(Conservation).count(), 0)
        conservation_values = pd.read_sql(session.query(Conservation).statement, session.bind)
        self.assertEqual(conservation_values.shape[1], 6)
        self.assertGreater(conservation_values.shape[0], 1000)
        self.assertEqual(conservation_values[conservation_values.conservation.isna()].shape[0], 0)
        self.assertEqual(conservation_values[conservation_values.conservation_sarbecovirus.isna()].shape[0], 0)
        self.assertEqual(conservation_values[conservation_values.conservation_vertebrates.isna()].shape[0], 0)



from unittest import TestCase
from dask.distributed import Client
from covigator.accessor.ena_accessor import EnaAccessor
from covigator.configuration import Configuration
from covigator.database.database import Database
from covigator.processor.ena_processor import EnaProcessor
from covigator.tests import SARS_COV_2_TAXID, HOMO_SAPIENS_TAXID
from covigator.dashboard.dashboard import Dashboard
import time


class IntegrationTests(TestCase):

    def setUp(self) -> None:
        self.config = Configuration()

    def test_access(self):
        """
        If given enough time, this creates a database with all runs complying with the selection criteria from ENA
        """
        accessor = EnaAccessor(tax_id=SARS_COV_2_TAXID, host_tax_id=HOMO_SAPIENS_TAXID,
                               database=Database(config=self.config))
        accessor.access()

    def test_process_dask(self):
        """
        If given enough time, this processes all ENA samples in the database
        """
        client = Client(n_workers=6, threads_per_worker=1)
        processor = EnaProcessor(database=Database(), dask_client=client, config=self.config)
        processor.process()

    def test_dashboard(self):
        dashboard = Dashboard(config=self.config)
        dashboard.start_dashboard(debug=True)
        time.sleep(60)



from unittest import TestCase

from covigator.misc.country_parser import CountryParser


class TestCountryParser(TestCase):

    def test_country_parser(self):
        parser = CountryParser()
        italy = parser.parse_country("iTAly")
        self.assertEqual(italy.country_raw, "iTAly")
        self.assertEqual(italy.country, "Italy")
        self.assertEqual(italy.country_alpha_2, "IT")
        self.assertEqual(italy.country_alpha_3, "ITA")
        self.assertEqual(italy.continent, "Europe")
        self.assertEqual(italy.continent_alpha_2, "EU")

        sofa = parser.parse_country("my sofa")
        self.assertEqual(sofa.country_raw, "my sofa")
        self.assertEqual(sofa.country, "Not available")
        self.assertEqual(sofa.country_alpha_2, "None")
        self.assertEqual(sofa.country_alpha_3, "None")
        self.assertEqual(sofa.continent, "None")
        self.assertEqual(sofa.continent_alpha_2, "None")

        empty = parser.parse_country("  ")
        self.assertEqual(empty.country_raw, "my sofa")
        self.assertEqual(empty.country, "Not available")
        self.assertEqual(empty.country_alpha_2, "None")
        self.assertEqual(empty.country_alpha_3, "None")
        self.assertEqual(empty.continent, "None")
        self.assertEqual(empty.continent_alpha_2, "None")



import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
from sqlalchemy.orm import Session
from dash.dependencies import Input, Output
import covigator
import covigator.configuration
from covigator.configuration import Configuration
from covigator.dashboard.tabs.dataset_ena import get_tab_dataset_ena
from covigator.dashboard.tabs.dataset_gisaid import get_tab_dataset_gisaid
from covigator.dashboard.tabs.footer import get_footer
from covigator.dashboard.tabs.overview import get_tab_overview
from covigator.dashboard.tabs.samples import get_tab_samples, set_callbacks_samples_tab
from covigator.dashboard.tabs.variants import get_tab_variants, set_callbacks_variants_tab
from covigator.database.database import get_database
from logzero import logger
from covigator.database.queries import Queries

VARIANTS_TAB_ID = "variants"
SAMPLES_TAB_ID = "samples"
GISAID_DATASET_TAB_ID = "gisaid-dataset"
ENA_DATASET_TAB_ID = "ena-dataset"
OVERVIEW_TAB_ID = "overview"


class Dashboard:

    def __init__(self, config: Configuration, verbose=False):
        covigator.configuration.initialise_logs(config.logfile_dash)
        self.config = config
        # the connection to the database is created only once, but multiple sessions are used
        self.database = get_database(config=config, initialize=True, verbose=verbose)

    def serve_layout(self):
        logger.info("Serving layout")
        footer = get_footer()
        layout = html.Div(children=[
            dbc.Navbar([

                # Use row and col to control vertical alignment of logo / brand
                dbc.Row(
                    [
                        dbc.Col(html.A(html.Img(
                            src="/assets/CoVigator_logo_txt_reg_no_bg.png", height="25px"),
                            href="https://covigator.tron-mainz.de/"), className="ml-2"
                        )
                    ],
                    align="center",
                    no_gutters=True,
                ),
                dbc.Row(
                    [
                        dbc.Col(html.A(html.Img(
                            src="/assets/tron_logo_without_text.png", height="22px"),
                            href="https://tron-mainz.de"), className="ml-2"),
                        dbc.Col(html.Br(), className="ml-2"),
                        dbc.Col(html.A(html.Img(
                            src="https://github.githubassets.com/images/modules/logos_page/Octocat.png", height="25px"),
                            href="https://github.com/TRON-bioinformatics/covigator"), className="ml-2"),
                        dbc.Col(html.Br(), className="ml-2"),

                    ],
                    align="center",
                    justify="end",
                    no_gutters=True,
                    style={'float': 'right', 'position': 'absolute', 'right': 0, 'text-align': 'right'}
                )
            ],
                color="white",
                dark=False,
            ),
            dbc.Card([
                dbc.CardHeader(
                    children=[
                        dbc.Tabs([
                            dbc.Tab(label="Overview", tab_id=OVERVIEW_TAB_ID),
                            dbc.Tab(label="ENA dataset", tab_id=ENA_DATASET_TAB_ID),
                            dbc.Tab(label="GISAID dataset", tab_id=GISAID_DATASET_TAB_ID),
                            dbc.Tab(label="Samples", tab_id=SAMPLES_TAB_ID),
                            dbc.Tab(label="Recurrent variants", tab_id=VARIANTS_TAB_ID)],
                            id="tabs",
                            active_tab="overview",
                            card=True),

                    ]),
                dbc.CardBody(dcc.Loading(id="loading-1", children=[html.Div(id="content")], type="default")),
                dbc.CardFooter(footer)
            ])
        ])
        return layout

    def start_dashboard(self, debug=False):
        try:
            logger.info("Starting covigator dashboard")
            app = self.get_application()
            app.run_server(debug=debug, host=self.config.dash_host, port=self.config.dash_port)
        except Exception as e:
            logger.exception(e)
            raise e

    def get_application(self) -> dash.Dash:

        # creates the Dash application
        logger.info("Create the application")
        app = dash.Dash(
            name=__name__,
            title="CoVigator",
            external_stylesheets=[dbc.themes.BOOTSTRAP],
            suppress_callback_exceptions=True,
            meta_tags=[
                # A description of the app, used by e.g.
                # search engines when displaying search results.
                {
                    'name': 'description',
                    'content': 'CoVigator - monitoring SARS-CoV-2 mutations'
                },
                # A tag that tells Internet Explorer (IE)
                # to use the latest renderer version available
                # to that browser (e.g. Edge)
                {
                    'http-equiv': 'X-UA-Compatible',
                    'content': 'IE=edge'
                },
                # A tag that tells the browser not to scale
                # desktop widths to fit mobile screens.
                # Sets the width of the viewport (browser)
                # to the width of the device, and the zoom level
                # (initial scale) to 1.
                #
                # Necessary for "true" mobile support.
                {
                  'name': 'viewport',
                  'content': 'width=device-width, initial-scale=1.0'
                }
            ]
        )
        # Warning pass the layout as a function, do not call it otherwise the application will serve a static dataset
        app.layout = self.serve_layout
        session = self.database.get_database_session()
        set_callbacks(app=app, session=session)
        set_callbacks_variants_tab(app=app, session=session)
        set_callbacks_samples_tab(app=app, session=session)
        return app


def set_callbacks(app, session: Session):

    queries = Queries(session=session)

    @app.callback(Output("content", "children"), [Input("tabs", "active_tab")])
    def switch_tab(at):
        logger.info("Changing tab...")
        try:
            if at == OVERVIEW_TAB_ID:
                return get_tab_overview(queries=queries)
            elif at == ENA_DATASET_TAB_ID:
                return get_tab_dataset_ena(queries=queries)
            elif at == GISAID_DATASET_TAB_ID:
                return get_tab_dataset_gisaid(queries=queries)
            elif at == SAMPLES_TAB_ID:
                return get_tab_samples(queries=queries)
            elif at == VARIANTS_TAB_ID:
                return get_tab_variants(queries=queries)
            return html.P("This shouldn't ever be displayed...")
        except Exception as e:
            logger.exception(e)


def main(debug=False):
    Dashboard(config=Configuration()).start_dashboard(debug=debug)


if __name__ == '__main__':
    main(debug=True)
